# try to reproduce results from Pollett2001, extended to 2D

import sys
import math
import time
#import scipy as np
import numpy as np
cimport numpy as np
import cython

import sir_stable

#import gaussiandict
include "gaussiandict.pyx"

"""
dI/dt = BSI - yI - uiI
i = I/N
=>
di/dt = dI/dt / N
      = BSI/N - yI/N - uiI/N
      = BN S/N I/N - yI/N - ui I/N
      = BNsi - yi - ui i

dS/dt = -BSI + u(N-S-I) + uiI
s = S/N
=>
ds/dt = dS/dt / N
      = -BSI/N + u(N-S-I)/N + uiI/N
      = -BN S/N I/N + u(N-S-I)/N + uiI/N
      = -BNsi + u(1-s-i) + ui i
"""

def my_sqrt(x):
    if x > 0:
        return math.sqrt(x)
    else:
        return 0

def signed_sqrt(x):
    if x >= 0:
        return math.sqrt(x)
    else:
        return -math.sqrt(-x)

# 4th order R-K method (Kreyszig, 8th Ed (1999) p.958
# If x = (s,i), then we are iteratilvely solving:
# dx = F(x) dt
cdef dx(double dt, double s, double i, double BN, double y, double u, double ui):
    cdef double dxs, dxi
    dxs = (-BN*s*i + u*(1-s-i) + ui*i) * dt
    dxi = (BN*s*i - y*i - ui*i) * dt
    return (dxs, dxi)

cdef detstep(double dt, double s, double i, double BN, double y, double u, double ui):
    cdef double k1s, k1i, k2s, k2i, k3s, k3i, k4s, k4i, snew, inew #, ds, di
    k1s, k1i = dx(dt, s, i, BN, y, u, ui)
    k2s, k2i = dx(dt, s + 0.5*k1s, i + 0.5*k1i, BN, y, u, ui)
    k3s, k3i = dx(dt, s + 0.5*k2s, i + 0.5*k2i, BN, y, u, ui)
    k4s, k4i = dx(dt, s + k3s, i + k3i, BN, y, u, ui)
    snew = s + (k1s + 2*k2s + 2*k3s + k4s) / 6.0
    inew = i + (k1i + 2*k2i + 2*k3i + k4i) / 6.0
    return (snew, inew)

# Note S0 and I0 need not be integers
cdef diffusion(int N, double BN, double y, double u, double ui, int S0, int I0,
               int Nfixed=1,
               int Istop=0, # exactly one of this and Sstop must be set
               int Smax=0, # additional Istop condition
               int verbose=0,
               op = None,
               # usually not used ones go below
               int timeit=0,
               int Sstop=0,
               comparison="<=", # usually what we want
               int Smin=0, # additional Istop condition
               int Imax=0, int Imin=0, # additional Sstop conditions
               ):
    cdef double B, s0, i0, tstep, t, tprint, s, i, smax, smin, imax, imin, sstop, istop
    cdef double Fs, Fi, Bss, Bsi, Bis, Bii, Gss, Gsi, Gis, Gii
    cdef int steps, Sd, Id
    if timeit:
        starttime = time.time()
    if op==None:
        op = sys.stdout
    if (Sstop and Istop) or (not Sstop and not Istop):
        raise Exception, "Exactly one of Sstop and Istop must be set"
    B = BN / float(N)
    s0 = S0 / float(N)
    i0 = I0 / float(N)
    steps = 0
    tstep = 0.001
    t = 0.0
    tprint = 5.0
    s = s0
    i = i0
    cdef np.ndarray[np.float64_t, ndim=2] V
    cdef np.ndarray[np.float64_t, ndim=2] Bu
    cdef np.ndarray[np.float64_t, ndim=2] BuT
    cdef np.ndarray[np.float64_t, ndim=2] G
    cdef np.ndarray[np.float64_t, ndim=2] dV
    V = np.zeros((2,2)) # (co)Variance matrix
    if Istop:
        istop = Istop / float(N)
        cstr = "i " + comparison + " istop"
        if Smax:
            smax = Smax / float(N)
            cstr += " and s <= smax"
        if Smin:
            smin = Smin / float(N)
            cstr += " and s >= smin"
        if verbose>=2 and N<=3000:
            op.write("cstr= " + str(cstr) + "\n")
            op.write("Istop= " + str(Istop) + "\n")
            op.write("istop= " + str(istop) + "\n")
            op.write("smax= " + str(smax) + "\n")
            op.write("Smax= " + str(smax*N) + "\n")
    else:
        sstop = Sstop / float(N)
        cstr = "s " + comparison + " sstop"
        if Imax:
            imax = Imax / float(N)
            cstr += " and i <= imax"
        if Imin:
            imin = Imin / float(N)
            cstr += " and i >= imin"
    while 1:
        if verbose>=3 and N<=3000:
            op.write("(S,I) = (%g,%g)\n" % (s*N, i*N))
        (s,i) = detstep(tstep, s, i, BN, y, u, ui)
        t += tstep
        steps += 1
        if i <= istop and s <= smax: #eval(cstr):
            if verbose>=2 and N<=3000:
                op.write("(S,I) = (%g,%g)\n" % (s*N, i*N))
            break
        if Nfixed:
            Fs = -BN*s*i + u - u*s - u*i + ui*i
            Fi =  BN*s*i - y*i - ui*i
            # Bss = dFs/ds = -BN*i - u
            # Bsi = dFs/di = -BN*s - u + ui
            # Bis = dFi/ds = BN*i
            # Bii = dFi/di = BN*s - y - ui
            Bss = -BN*i - u
            Bsi = -BN*s - u + ui
            Bis =  BN*i
            Bii =  BN*s - y - ui
            # g11 = (-1)(-1)BNsi + (0)(0)yi  + (1)(1)u(1-s-i) + (1)(1)ui i
            # g12 = (-1)( 1)BNsi + (0)(-1)yi + (1)(0)u(1-s-i) + (1)(-1)ui i
            # g21 = ( 1)(-1)BNsi + (-1)(0)yi + (0)(1)u(1-s-i) + (-1)(1)ui i
            # g22 = ( 1)( 1)BNsi + (-1)(-1)yi + (0)(0)u(1-s-i) + (-1)(-1)ui i
            Gss =  BN*s*i + u*(1-s-i) + ui*i
            Gsi = -BN*s*i - ui*i
            Gis = -BN*s*i - ui*i
            Gii =  BN*s*i + y*i + ui*i
        else:
            Fs = -BN*s*i + u - u*s
            Fi =  BN*s*i - (y+u)*i
            # Bss = dFs/ds = -BN*i - u
            # Bsi = dFs/di = -BN*s
            # Bis = dFi/ds = BN*i
            # Bii = dFi/di = BN*s - (y+u)
            Bss = -BN*i - u
            Bsi = -BN*s
            Bis =  BN*i
            Bii =  BN*s - (y+u)
            # g11 = (-1)(-1)BNsi + (0)(0)(y+u)i  + (1)(1)u + (-1)(-1)us
            # g12 = (-1)( 1)BNsi + (0)(-1)(y+u)i + (1)(0)u + (-1)(0)us
            # g21 = ( 1)(-1)BNsi + (-1)(0)(y+u)i + (0)(1)u + (0)(-1)us
            # g22 = ( 1)( 1)BNsi + (-1)(-1)(y+u)i + (0)(0)u + (0)(0)us
            Gss =  BN*s*i + u*(1+s)
            Gsi = -BN*s*i
            Gis = -BN*s*i
            Gii =  BN*s*i + (y+u)*i
            
        Bu = np.matrix([[Bss, Bsi],
                        [Bis, Bii]])
        BuT = np.transpose(Bu)
        G = np.matrix([[Gss, Gsi],
                       [Gis, Gii]])
        dV = (Bu*V + V*BuT + G) * tstep
        V = V + dV
        if verbose >= 2 and t > tprint:
            op.write("t=%.1f, det=(%.2f, %.2f), sd = (%.2f, %.2f, %.2f)\n" %
                             (t, s*N, i*N,
                              my_sqrt(V[(0,0)]*N),
                              my_sqrt(V[(1,1)]*N),
                              signed_sqrt(V[(0,1)]*N)))
            tprint += 5.0
    # end of loop. Little to be gained from cython defs after here
    if verbose >= 2:
        op.write("exit time t = %g\n" % t)
    # use final values of Fs and Fi
    if Istop:
        hitvar0 = V[0,0] - (Fs/Fi) * V[1,1]
        hitvar1 = V[0,0] + (Fs/Fi)**2 * V[1,1]
        hitvar  = V[0,0] + (Fs/Fi)**2 * V[1,1] - 2*(Fs/Fi)*V[0,1]
    else:
        hitvar0 = V[1,1] - (Fi/Fs) * V[0,0]
        hitvar1 = V[1,1] + (Fi/Fs)**2 * V[0,0]
        hitvar  = V[1,1] + (Fi/Fs)**2 * V[0,0] - 2*(Fi/Fs)*V[0,1]
    hitsd0 = my_sqrt(hitvar0) / math.sqrt(N)
    hitsd1 = my_sqrt(hitvar1) / math.sqrt(N)
    hitsd  = my_sqrt(hitvar)  / math.sqrt(N)
    if verbose >= 1:
        if Istop:
            stopstr = "Istop"
        else:
            stopstr = "Sstop"
        op.write("%s diffusion stddevs at (%.2f, %.2f) are (%.2f, %.2f, %.2f) -> %.1f (old: %.1f, %.1f)\n"
                         % (stopstr, s*N, i*N, my_sqrt(V[0,0]*N), my_sqrt(V[1,1]*N), signed_sqrt(V[0,1]*N), hitsd*N, hitsd1*N, hitsd0*N))
        if Istop:
            # discretise
            (Sd, Id) = sir_stable.stable(N, BN, y, u, ui)
            Sdict = gaussiandict(s*N, hitsd*N, 0, Sd-1, rescale=1)
            dmean = 0.0
            dmean2 = 0.0
            dtot = 0.0
            for key in Sdict:
                dtot   += Sdict[key]
                dmean  += Sdict[key] * key
                dmean2 += Sdict[key] * key*key
            dmean = dmean / dtot
            dmean2 = dmean2 / dtot
            dsd = my_sqrt(dmean2 - dmean*dmean)
            op.write("After discretisation (of dtot %.3f ),            mean %.1f and sd %.1f\n" % (dtot, dmean, dsd))
    if timeit:
        op.write("elapsed time = %g seconds (diffusion.pyx, %d steps)\n" % (time.time() - starttime, steps))
    if Istop:
        return (s*N, hitsd*N, t, i*N, Fs/Fi)
    else:
        return (i*N, hitsd*N, t, s*N, Fi/Fs)
