# returns a dictionary filled with a discretised gaussian distribution

import sys
import math
import time
import scipy.stats

# mean and sd may be float; imin and imax must be int
# if rescale=1, final probs sum to 1
# else, out of range prob is simply discarded
cdef gaussiandict(double mean, double sd, int imin, int imax, int rescale=1, int verbose=0, int timeit=0):
    cdef int i, centre
    cdef double xtot, imean, imean2, scale, x, v, denom, leftover, ivar, isd
    if timeit:
        starttime = time.time()
    if mean > imax:
        sys.stdout.write("WARNING imax (%d) < mean (%d)\n" % (imax, mean))
    if mean < imin:
        sys.stdout.write("WARNING mean (%d) < imin (%d)\n" % (mean, imin))
    #if imax - mean < mean - imin:
    #    usemax = imax
    #    usemin = mean - (imax-mean)
    #else:
    #    usemin = imin
    #    usemax = mean + (mean-imin)

    odict = {}
    xtot = 0.0
    imean = 0.0
    imean2 = 0.0
    if rescale:
        # apply scaling to distribute out-of-range prob mass evenly
        outofrange_left = scipy.stats.norm.cdf((imin-mean-0.5)/sd)
        outofrange_right = 1.0 - scipy.stats.norm.cdf((imax-mean+0.5)/sd)
        scale = 1.0 / (1.0 - outofrange_left - outofrange_right)
        if verbose:
            sys.stdout.write("mean=%s, sd=%s, imin=%d, imax=%d\n" % (mean, sd, imin, imax))
            sys.stdout.write("OOR_l=%s, OOR_r=%s, scale=%s\n" % (outofrange_left, outofrange_right, scale))
    else:
        sys.stdout.write("mean=%s, sd=%s, imin=%d, imax=%d (no rescaling)\n" % (mean, sd, imin, imax))
        scale = 1.0
    #if timeit:
    #    sys.stdout.write("time before loop = %g seconds (gaussiandict.py)\n" % (time.time() - starttime))
    if sd <= 0.0:
        for i in range(imin, imax+1):
            odict[i] = 0.0
    else:
        denom = scale / (math.sqrt(2*math.pi) * sd)   # actually 1/denom
        for i in range(imin, imax+1):
            v = (i-mean)/sd
            x = math.exp(-v*v/2.0) * denom
            # the scipy calls seem to take all the time
            #x = scipy.stats.norm.pdf(mean) # see if this alone slows it down
            #x = scale / (imax+1.0-imin)
            #x = scale * (scipy.stats.norm.cdf((i-mean+0.5)/sd) - scipy.stats.norm.cdf((i-mean-0.5)/sd))
            #x = scale * scipy.stats.norm.pdf((i-mean)/sd) / sd
            odict[i] = x
            xtot += x
            imean += x * i
            imean2 += x * i*i
    #if timeit:
    #    sys.stdout.write("time after loop = %g seconds (gaussiandict.py, %d steps)\n" % (time.time() - starttime, imax+1-imin))
    if rescale:
        # add any leftover to centre
        leftover = 1.0 - xtot
        if verbose:
            sys.stdout.write("gaussiandict leftover = %g\n" % leftover)
        if sd > 0.0 and abs(leftover) > 1e-3:
            if verbose >= 3:
                for i in range(imin, imax+1):
                    sys.stderr.write("dict[%d] = %g\n" % (i, odict[i]))
            raise Exception, "unexpectedly large leftover %g" % leftover
        centre = int(mean+0.5)
        if centre not in odict:
            odict[centre] = 0.0
        imean2 -= odict[centre] * centre*centre
        odict[centre] += leftover
        imean += centre * leftover
        imean2 += odict[centre] * centre*centre
    else:
        imean = imean / xtot
        imean2 = imean2 / xtot
    ivar = imean2 - imean*imean
    isd = math.sqrt(max(ivar, 0))
    if verbose:
        sys.stdout.write("Result dist has mean=%.1f and sd=%.1f\n" % (imean, isd))
    if timeit:
        sys.stdout.write("elapsed time = %g seconds (gaussiandict.py)\n" % (time.time() - starttime))
    return odict
