# Taken from 
# Computation of epidemic final size distributions
# Andrew J. Black and J. V. Ross
# School of Mathematical Sciences, The University of Adelaide, Adelaide, SA 5005, Australia.
# Submitted to Proceedings of the National Academy of Sciences of the United States of America
# March 2014
#
# Modified by PB to do phases (29/5/14 onwards):
# Fri Mar 20 08:46:32 CST 2015 - Cython (markovfull.pyx) from blackross_sirs.jl
#
import sys
import math
import time
import os
import numpy as np
cimport numpy as np
#import cython

import sir_stable

# Example for Nlimit=6
#   6 | 27
#   5 | 25  26
#   4 | 22  23  24
#   3 | 18  19  20  21
#   2 | 13  14  15  16  17 
#   1 |  7   8   9  10  11  12
# I=0 |  0   1   2   3   4   5   6
# ------------------------------------
#  S =   0   1   2   3   4   5   6  ...
#
# Infection         (S-1,I+1) i + (Nlimit-I)
# Recovery          (S,  I-1) i - (Nlimit-I+2)
# S birth / R renew (S+1,I)   i + 1  ... and S+I+1 may exceed Nlimit
# I renewal         (S+1,I-1) i - (Nlimit-I+1)
# S death           (S-1,I)   i - 1

cdef markovfull(ofile,
                int N, double BN, double y, double u, double ui, int I0,
                int Sd, int Id, int Iupper2,
                int Nfixed, int phase_method, int timeit, verbose=1, int S0=-1):
    cdef int max_Z3, Nlimit, K, phases, phase, DUMP, i, Z3, Z2, Z1, S, I, R, jump12, jump23, jump34, phasejump, j
    cdef double B, val, a1, a2, a3, a4, tot, pleft, p0, p1, pinit, plimit, rem
    if timeit:
        starttime = time.time()
    if ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("markovfull.pyx"):
            #sys.stdout.write("markovfull Up to date!\n")
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")

    if S0 < 0:
        S0 = N - I0
    B = BN / float(N)

    max_Z3 = 100 * N
    # Nlimit is the hard boundary, equal to N if Nfixed
    if Nfixed:
        Nlimit = N
    else:
        Nlimit = int(N * 1.1 + 0.5)
    K = int((Nlimit+1)*(Nlimit+2)/2)
    phases = 2
    op.write("markovfull: N=%d, BN=%s, y=%s, u=%s, ui=%s, I0=%d, phase_method=%d, Sd=%d, Id=%d, Nlimit=%d\n"
                     % (N, BN, y, u, ui, I0, phase_method, Sd, Id, Nlimit))
    
    #cdef np.ndarray[np.float64_t, ndim=2] abs_probs
    #cdef np.ndarray[np.float64_t, ndim=1] final_size
    cdef np.ndarray[np.float64_t, ndim=2] p_vec
    cdef np.ndarray[np.float64_t, ndim=2] h_vec
    #cdef np.ndarray[np.float64_t, ndim=1] prob_phase
    #cdef np.ndarray[np.float64_t, ndim=1] subtot

    #abs_probs = np.zeros((Nlimit+2,phases))
    #final_size = np.zeros((phases))
    p_vec = np.zeros((K+1,phases))
    h_vec = np.zeros((K+1,phases))
    pinit = 0.0
    p0 = 0.0
    p1 = 0.0
    plimit = 0.0
    rem = 1e-6 # exit when unallocated prob (pleft) < rem

    i = 0
    for I in range(Nlimit+1):
        for S in range(Nlimit+1-I):
            if S==S0 and I==I0:
                p_vec[i,0] = 1.0
                op.write("Initial point (%d,%d), p_vec[%d,0]=1\n" % (S, I, i))
                break
            i = i+1
                                                                           
    #prob_phase = np.zeros(phases)
    DUMP = 0
    for Z3 in range(max_Z3+1):
        i = 0
        for I in range(Nlimit+1):
            for S in range(Nlimit+1-I):
                R = N - S - I
                a1 = B*S*I
                if Nfixed:
                    a2 = y*I
                    a3 = max(u*R, 0.0)
                    a4 = ui*I
                    a5 = 0.0
                else:
                    a2 = (y+ui)*I
                    a3 = u*N
                    a4 = 0.0
                    a5 = u*S
                tot = a1+a2+a3+a4+a5 # will be float
                if I>0 and tot > 0.0:
                    for phase in range(phases):
                        if verbose>3 and p_vec[(i,phase)]>0.0:
                            op.write("Z3=%d, i=%d, (S,I)=(%d,%d), phase=%d, p_vec[i,phase]=%.6f: %.4f, %.4f, %.4f, %.4f\n"
                                             % (Z3, i, S, I, phase, p_vec[(i,phase)], a1/tot, a2/tot, a3/tot, a4/tot))
                        # S,I -> S-1,I+1 (infection); phase 1->2 or 3->4 if I rising
                        if phase==1 and (S-1>=Sd and I+1>=Id or I+1>=Iupper2):
                            p1 += p_vec[i,phase]*a1/tot
                        elif phase==0 and (S-1<Sd and I+1==Id):
                            h_vec[i+Nlimit-I,phase+1] += p_vec[i,phase]*a1/tot
                        else:
                            h_vec[i+Nlimit-I,phase]   += p_vec[i,phase]*a1/tot

                        # S,I -> S,I-1   (recovery); phase 2->3 if I falling
                        if phase==1 and (S>=Sd and I-1>=Id or I-1>=Iupper2):
                            p1 += p_vec[i,phase]*a2/tot
                        elif phase==1 and I-1==0:
                            p0 += p_vec[i,phase]*a2/tot
                        elif phase==0 and I-1==0:
                            pinit += p_vec[i,phase]*a2/tot
                        elif phase==0 and S<Sd and I-1==Id:
                            h_vec[i-(Nlimit-I+2),phase+1] += p_vec[i,phase]*a2/tot
                        else:
                            h_vec[i-(Nlimit-I+2),phase]   += p_vec[i,phase]*a2/tot

                        if a3>0:
                            # S,I -> S+1,I   (r renewal or s birth, jump to the next plane)
                            if S+I+1>Nlimit:
                                plimit += p_vec[i,phase]*a3/tot
                            elif phase==1 and (S+1>=Sd and I>=Id or I>=Iupper2):
                                p1 += p_vec[i,phase]*a3/tot
                            elif phase==0 and S+1<Sd and I==Id:
                                h_vec[i+1,phase+1] += p_vec[i,phase]*a3/tot
                            else:
                                h_vec[i+1,phase]   += p_vec[i,phase]*a3/tot

                        # S,I -> S+1,I-1   (I renewal, jump to next plane)
                        if phase==1 and (S+1>=Sd and I-1>=Id or I-1>=Iupper2):
                            p1 += p_vec[i,phase]*a4/tot
                        elif phase==1 and I-1==0:
                            p0 += p_vec[i,phase]*a4/tot
                        elif phase==0 and I-1==0:
                            pinit += p_vec[i,phase]*a4/tot
                        elif phase==0 and S+1<Sd and I-1==Id:
                            h_vec[i-(Nlimit-I+1),phase+1] += p_vec[i,phase]*a4/tot
                        else:
                            h_vec[i-(Nlimit-I+1),phase]   += p_vec[i,phase]*a4/tot
                            
                        if a5>0:
                            # S,I -> S-1,I   (s death)
                            if phase==1 and (S-1>=Sd and I>=Id or I>=Iupper2):
                                p1 += p_vec[i,phase]*a5/tot
                            elif phase==0 and S-1<Sd and I==Id:
                                h_vec[i-1,phase+1] += p_vec[i,phase]*a5/tot
                            else:
                                h_vec[i-1,phase]   += p_vec[i,phase]*a5/tot
                i = i+1
        p_vec = h_vec
        h_vec = np.zeros((K+1,phases))

        i = 0
        pleft = 0.0
        for I in range(Nlimit+1):
            for S in range(Nlimit+1-I):
                for phase in range(phases):
                    pleft += p_vec[(i,phase)]
                i = i + 1

        if verbose>1 and Z3%100 == 0 or verbose>2:
            op.write("Z3=%d\t%.12f + %.12f + %.12f + %.12f + %.12f = %.12f\n"
                             % (Z3, pleft, pinit, p0, p1, plimit, pleft+pinit+p0+p1+plimit))
        if pleft < rem:
            break
        if verbose>3 and Z3>=2:
            break
    else:
        sys.stderr("Warning, pleft=%.6f > %.6f, result may be unreliable\n" % (pleft, rem))

    if verbose:
        op.write("exit: Z3=%d\t%.12f + %.12f + %.12f + %.12f + %.12f = %.12f\n"
                 % (Z3, pleft, pinit, p0, p1, plimit, pleft+pinit+p0+p1+plimit))

    op.write("p0=%.6f  (raw: pinit=%.6f, p0=%.6f, p1=%.6f, plimit=%.6f, rem=%.6f)\n"
                     % (p0 / (p0+p1), pinit, p0, p1, plimit, rem))

    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if op!="":
        op.close()
        os.rename(ofile+"fail", ofile)
