# Simplified algorithm, modifying markov2bound.jl

#Pkg.add("PyPlot")
#using PyPlot

import sys
import math
import time
import os
import numpy as np
cimport numpy as np
import cython
#import matplotlib.pyplot as plt

import sirdetpeaks       # deterministic solution
#import sirplot

# This has absorbing boundaries at I=0 and I=Icolmax
# In the example below Icolmax=6. Note the row "step" is Icolmax-1
#
#   5 |  1   6  11  16  21  26  31  ...
#   4 |  2   7  12  17  22  27  32  ...
#   3 |  3   8  13  18  23  28  33  ...
#   2 |  4   9  14  19  24  29  34  ...
# I=1 |  5  10  15  20  25  30  35  ...
# ------------------------------------
#  S =   0   1   2   3   4   5   6  ...
#
# a   Infection-Renew pair, up (I+) at rate BSI
# b   Recovery, down (I-) at rate yI
# g   Excess Renew, right (S+) at rate max(uR - BSI, 0)
#  also add ui*I as two separate events, one down (add to b), one right (add to c)

# It seems defaulted keywords can only be omitted "in reverse order",
# i.e. if you want to redefine a param, you need to define all the ones before it.
# But at calling function can put them in any order
cdef sweeptrough(ofile,
                 int N, double BN, double y, double u, double ui,
                 int Sd, int Iupper, int Iupper2,
                 int monitor_S = 0, int Nfixed = 1, int timeit = 0, int verbose = 1,
                 startdict = None,
                 int method = 1,
                 double hratio = 0.5, # use for method=9 only
                 # usually don't declare these, just use defaults
                 double rem = 1e-6,
                 double brate = -1.0, double sdeath = -1.0):
    cdef int Icolmax, S, I, R, Sstop, count, count1, count9, j
    cdef double B, Dmoment, Dmass, Iav, denom, Dsum, p1, p0, ratio, BSI, yI, uR, uiI, huR
    # verbose=0 for silent, 1 for about 3 lines of output, 2 for debug messages
    # monitor_S: if set, we do not exit region until I>=Ie AND S>=Sd 
    #    i.e. not until there is "upward pressure" on S.
    #    But we limit I to Iupper2, and if it happens to reach it we say we've hit upper bound anyway
    # rem = unallocated probability, i.e. terminate when only rem of prob has failed to hit a boundary
    if timeit:
        starttime = time.time()
    if type(ofile)==file:
        op = ofile
    elif ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("sweeptrough.pyx"):
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")
    B = BN / float(N)
    if not Nfixed and brate < 0:
        brate = u*N # note this is the fixed N value (i.e. a fixed birth rate)
    if not Nfixed and sdeath < 0:
        sdeath = u
    if method==1:
        methodstr = ""
    elif method==9:
        methodstr = "(%d_%.1f)" % (method, hratio)
    else:
        methodstr = "(%d)" % method
    if verbose >= 1:
        op.write("sweeptrough%s: " % methodstr)
        op.write("N=%d, BN=%s, y=%s, u=%s, ui=%s, monitor_S=%s, Nfixed=%s\n"
                          % (N, BN, y, u, ui, monitor_S, Nfixed))

    # point 5 = Imin (float) - for information (trough depth) only
    # *** This belongs in a different program - unrelated to sweep algorith
    #(Sdf, Imin, tignore, Sdiffignore, Idiffignore) = sirdetpeaks.detpoint(N=N, BN=BN, y=y, u=u, ui=ui, I0=Istart, point=5, forceint=False)

    count = 0
    count1 = 0
    count9 = 0
    p0 = 0.0 # cumulative prob of absorption at lower bound (I=0)
    p1 = 0.0 # cumulative prob of absorption at upper bound (I=Icolmax)
    cdef np.ndarray[np.float64_t, ndim=1] D
    cdef np.ndarray[np.float64_t, ndim=1] a
    cdef np.ndarray[np.float64_t, ndim=1] b
    cdef np.ndarray[np.float64_t, ndim=1] bprev
    cdef np.ndarray[np.float64_t, ndim=1] f
    cdef np.ndarray[np.float64_t, ndim=1] g
    cdef np.ndarray[np.float64_t, ndim=1] h
    cdef np.ndarray[np.float64_t, ndim=1] r
    cdef np.ndarray[np.float64_t, ndim=1] V
    cdef np.ndarray[np.float64_t, ndim=1] W
    cdef np.ndarray[np.float64_t, ndim=1] T
    Sstop = 0
    total_I_error = 0.0
    total_I_abserror = 0.0
    if startdict:
        allocated = 0.0
        for key in startdict.keys():
            allocated += startdict[key]
        op.write("sweeptrough beginning with %.6f prob mass in startdict{}\n" % allocated)
    for S in range(0, N):
        if S < Sd and monitor_S:
            Icolmax = Iupper2
        else:
            Icolmax = Iupper
        if S==0:
            D = np.zeros(Icolmax) # Distribution (of prob mass). Column at given S
        for I in range(1, Icolmax):
            if (S,I) in startdict:
                D[I] += startdict[(S,I)]
        a = np.zeros((Icolmax)) # up rate
        b = np.zeros((Icolmax)) # down rate
        f = np.zeros((Icolmax)) # up-right rate
        g = np.zeros((Icolmax)) # right rate
        h = np.zeros((Icolmax)) # down-right rate
        r = np.zeros((Icolmax)) # total rate
        if verbose>=2 and method==2 and S>0 and S<Sd:
            format = "%" + str(int(len(str(N)))) + "d:"
            op.write(format % S)
        for I in range(1, Icolmax): # i.e. 1 to N inclusive, or 1 to Iupper-1 inclusive
            R = N - S - I
            BSI = B*S*I
            yI = y*I
            uR = u*R
            uiI = ui*I
            count += 1
            if method==9:
                huR = hratio * uR
            # method==1 is default, others are experimental
            if (method==1
                or method==9 and (BSI > huR or huR > 2*BSI or huR > yI)
                or method==10 and (BSI > yI or 2*BSI > uR)):
                if I == Icolmax-1:
                    f[I] = BSI # absorbing at I==Icolmax
                else:
                    a[I] = BSI
                if Nfixed:
                    if uR + uiI > BSI:
                        g[I] = uR + uiI - BSI
                        count1 += 1
                    else:
                        g[I] = 0.0
                else: # in Nfixed==0, uiI is a down transition so not included here. 7/7/15
                    g[I] = max(brate - u*S - BSI, 0.0)
                if I > 1:
                    b[I] = yI + uiI
                else:
                    h[I] = yI + uiI # absorbing at I==0
                r[I] = a[I] + b[I] + f[I] + g[I] + h[I]
            elif method==9:
                if I == Icolmax-1:
                    f[I] = BSI # a and f both, absorbing at I==Icolmax
                else:
                    a[I] = 2*BSI - huR
                    f[I] = huR - BSI
                g[I] = uR - 2*huR
                if I > 1:
                    b[I] = yI - huR
                    h[I] = huR
                else:
                    h[I] = yI # b and h both absorbing at I==0
                r[I] = a[I] + b[I] + f[I] + g[I] + h[I]
                count9 += 1
            elif method==10:
                if I == Icolmax-1:
                    f[I] = BSI
                else:
                    a[I] = BSI
                g[I] = uR - 2*BSI
                if I > 1:
                    b[I] = yI - BSI
                    h[I] = BSI
                else:
                    h[I] = yI # b and h both absorbing at I==0
                r[I] = a[I] + b[I] + f[I] + g[I] + h[I]
                count9 += 1
            elif method <= 3 or method==6 and uR <= BSI: 
                # method 2 will not work if S is too large,
                # because get to point where deterministic curve never reaches S+1
                if method<=1 or method==6 or method==2 and (S==0 or S>=Sd):
                    a[I] = BSI
                    if Nfixed:
                        g[I] = max(uR + uiI - BSI, 0.0)
                    else: # note this is the same when u==ui and R+I==N-S
                        g[I] = max(brate - u*S + uiI - BSI, 0.0)
                elif method==2:
                    # deterministic run from end point of BSI event (S-1, I+1)
                    # back to current S value
                    (Sdet, Idet, tdet) = sirdetpeaks.sirdetrun(N=N, BN=BN, y=y, u=u, ui=ui,
                                                               S0=S-1, I0=I+1, Sstop=S)
                    ratio = Idet - I
                    if ratio < 0 or ratio > 1:
                        #op.write("warning ratio=%g: (%d,%d) -> (%g,%g)\n"
                        #                 % (ratio, S-1, I+1, Sdet, Idet))
                        if ratio < 0:
                            a[I] = 0.0
                        else:
                            a[I] = BSI
                    else:
                        a[I] = BSI * ratio
                    g[I] = uR + uiI # don't subtract BSI if method==2
                    if verbose >= 2:
                        op.write(" (%d) %.4f" % (I, ratio))
                else:
                    if uR == BSI:
                        a[I] = 100.0 * BSI
                    else:
                        a[I] = BSI * (uR - yI) / float(uR - BSI)
                    g[I] = uR + uiI # don't subtract BSI if method==3
                b[I] = yI + uiI
                if method==0:
                    if S>0 and g[I]>0:
                        b[I] = b[I] * (1.0 - a[I]/float(g[I])) + bprev[I]*a[I]/float(g[I])
                    bprev = np.copy(b)
                r[I] = a[I] + b[I] + g[I]
                if verbose>=2 and method==2 and S>0 and S<Sd:
                    op.write("\n")
            elif method==4: # meeeting, 10-June (Josh)
                a[I] = 1.5*BSI
                b[I] = yI + 0.5*BSI + uiI
                g[I] = max(uR + uiI - BSI, 0.0)
                r[I] = a[I] + b[I] + g[I]
            elif method==5: # Josh 11-June, #1 note: currently assumes ui==0
                denom = BSI + yI + uR
                a[I] = BSI * (2*BSI + yI) / float(denom * (BSI + yI))
                b[I] = yI * (2*BSI + yI) / float(denom * (BSI + yI))
                g[I] = max(uR - BSI, 0.0) / float(denom)
                r[I] = a[I] + b[I] + g[I]
            elif method==6: # meeting, 10-June (me)
                a[I] = 0.5*BSI * (2*uR + uiI - BSI - yI) / float(uR + uiI - BSI)
                b[I] = yI + 0.5*BSI * (yI + uiI - BSI) / float(uR + uiI - BSI)
                g[I] = uR + uiI
                r[I] = a[I] + b[I] + g[I]
            elif method==7: # Josh 11-June, #2 note: currently assumes ui==0
                a[I] = BSI * max(uR - yI, 0)
                b[I] = (uR + BSI) * max(uR - BSI, 0)
                g[I] = max(uR * yI - (BSI)**2, 0)
                r[I] = a[I] + b[I] + g[I]
            elif method==8: # balance ratio of drifts and diffusions, but not I components (SIRS only)
                if yI * uR > BSI * BSI:
                    a[I] = max(BSI*yI*(uR-yI) / float(yI*uR - BSI*BSI), 0)
                    b[I] = yI
                else: # BSI denom is zero or neg, so make a >>>>> b
                    a[I] = BSI
                    b[I] = 0
                g[I] = (BSI+uR)*(yI + a[I]) / float(BSI+yI)
                r[I] = a[I] + b[I] + g[I]
            else:
                raise Exception, "Unknown method " + str(method)
                
        # see ../doc/sweep.txt for explanation of these next 2 for loops
        # V and W satisfy: T[I-1] = V[I-1] + W[I-1]*T[I]
        V = np.zeros((Icolmax))
        W = np.zeros((Icolmax))
        T = np.zeros((Icolmax)) # expected time in each state (assuming one step per time unit)
        for I in range(1, Icolmax):
            if I==1:
                # T[1] = D[1] + T[2]*b[2]/r[2]
                # => V[1] = D[1], W[1] = b[2]/r[2]
                V[I] = D[I]
                W[I] = b[I+1] / float(r[I+1])
            else:
                # T[I] = D[I] + T[I+1]*b[I+1]/r[I+1] + T[I-1]*a[I-1]/r[I-1]
                #      = D[I] + T[I+1]*b[I+1]/r[I+1] + (V[I-1] + W[I-1]*T[I])*a[I-1]/r[I-1]
                # => (1 - W[I-1]*a[I-1]/r[I-1])*T[I]  = (D[I] + V[I-1]*a[I-1]/r[I-1]) + T[I+1]*b[I+1]/r[I+1]
                #  so if we set demon = 1 - W[I-1]*a[I-1]/r[I-1]
                # => T[I]  = (D[I] + V[I-1]*a[I-1]/r[I-1])/denom + T[I+1]*b[I+1]/r[I+1]/denom
                # =>    V[I] = (D[I] + V[I-1]*a[I-1]/r[I-1])/denom
                #  and  W[I] = b[I+1]/r[I+1] / denom
                denom = 1.0 - a[I-1] / float(r[I-1]) * W[I-1]
                V[I] = (D[I] + a[I-1] / float(r[I-1]) * V[I-1]) / denom
                if I < Icolmax-1:
                    W[I] = (b[I+1] / float(r[I+1])) / denom
                else:
                    W[I] = 0.0
        # Now simply go back down the line, evaluating T[I] = V[I] + W[I]*T[I+1]
        for I in range(Icolmax-1, 0, -1):
            if I==Icolmax-1:
                T[I] = V[I]
            else:
                T[I] = V[I] + W[I] * T[I+1]
        D = np.zeros((Icolmax)) # Distribution of probability mass
        Dsum = 0.0
        for I in range(Icolmax-1, 0, -1): # Calculate new D values
            if S==Sd-1 and I>=Iupper:
                # special case where Icolmax is lower on next step
                p1 += T[I] * (f[I] + g[I]) / float(r[I])
                if I-1 >= Iupper:
                    p1     += T[I] * h[I] / float(r[I])
                else:
                    D[I-1] += T[I] * h[I] / float(r[I])
            else:
                if I == Icolmax-1:
                    p1     += T[I] * f[I] / float(r[I])
                else:
                    D[I+1] += T[I] * f[I] / float(r[I])
                D[I]       += T[I] * g[I] / float(r[I])
                if I > 1:
                    D[I-1] += T[I] * h[I] / float(r[I])
                else:
                    p0     += T[I] * h[I] / float(r[I])
            Dsum += D[I]
        if verbose >= 4:
            op.write("D (1..%d) = [" % (Icolmax-1))
            for j in range(1, Icolmax):
                op.write("%.4f, " % D[j])
            op.write("]\n")
        if verbose >= 2:
            if verbose >=3 or S % 1000 == 0:
                op.write("S=%d, p0=%.8f, p1=%.8f, Dsum=%.8f, total=%.8f\n"
                                 % (S, p0, p1, Dsum, p0+p1+Dsum))
        Sstop = S
        if 1 - p0 - p1 < rem:
            break
        if 0: #not (0.99 < p0+p1+Dsum < 1.01):
            op.write("Aborting because something is wrong: p0+p1+Dsum = %s but should be 1\n" % (p0+p1+Dsum))
            op.write("a = %s\n" % a)
            op.write("b = %s\n" % b)
            op.write("g = %s\n" % g)
            op.write("r = %s\n" % r)
            #printmatrix(X % "X")
            op.write("D = %s\n" % D)
            break
    if verbose >= 1:
        if method>=9:
            op.write("%d : %d / %d points used method %d : 1\n" % (count9, count1, count, method))
            s = "_%s" % hratio
        else:
            s = ""
        op.write("sweeptrough%s: p0=%.6f  p1=%.6f  rem=%f  Sstop=%d (method=%d%s)\n"
                         % (methodstr, p0, p1, 1 - p0 - p1, Sstop, method, s))
        #p0_true_predict = p0 / predicted_p0_ratio # i.e expect true p0 to be a little higher
        #op.write("predict true p0 between %g and %g\n" % (p0, p0_true_predict))
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if type(ofile)!=file and op!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return p1
