import sys
import os
import time
import matplotlib.pyplot as plt
import numpy as np

import sirdefs
import getpolicy
import markovfull
import sirplot

def printpolicy(policy, name):
    Ilist = []
    Slist = {}
    for (S,I) in policy:
        if I not in Ilist:
            Ilist.append(I)
            Slist[I] = [S]
        else:
            Slist[I].append(S)
    Ilist.sort()
    sys.stdout.write("policy %s:\n" % name)
    for I in Ilist:
        sys.stdout.write("I=%d:" % I)
        Slist[I].sort()
        for S in Slist[I]:
            sys.stdout.write(" %d" % S)
        sys.stdout.write(" (%d)\n" % len(Slist[I]))

def plotapprox(N, B1, B2, y, u1, u2, u3, Nfixed, color):
    (Sd1, Id1) = sirdefs.stable(      N, B1, y, u1, u2, u3)
    (Se1, Ie1) = sirdefs.stable_float(N, B1, y, u1, u2, u3)
    (Sd, Id)   = sirdefs.stable(      N, B2, y, u1, u2, u3)
    (Se, Ie)   = sirdefs.stable_float(N, B2, y, u1, u2, u3)
    #sys.stdout.write("(Se,Ie)=(%.3f,%.3f)  (Sd,Id)=(%d,%d)\n" % (Se, Ie, Sd, Id))
    Slist = []
    Ilist = []
    for S in range(N):
        Slist.append(S)
        if S < Sd1:
            Ilist.append( max((N-S) * (u1+u2) / float(y+u1+u2), Id1))
        elif S < Sd:
            Ilist.append(Id1)
        else:
            if Nfixed:
                Ilist.append(min(Id, N-S))
            else:
                Ilist.append(Id)
    plt.plot( Slist, Ilist, color=color, label="simplified policy") #, linewidth=2)
    ymax = max(Ilist)
    plt.ylim(ymin=0, ymax=ymax)

def plotpolicy(policy, rempolicy, N, B1, y, u1, u2, u3, color, linestyle, label, Ipolicytop):
    scatter = 0
    contour = 0
    ymax = 0
    if scatter:
        for (S,I) in policy | rempolicy:
            plt.scatter( S, I, color=color, s=5)
            ymax = max(I, ymax)
        plt.ylim(ymin=0, ymax=ymax)
        #plt.ylim(ymin=0, ymax=30)
    elif contour:
        # leave room for the legend above any of the contours
        # 1.7 chosen empirically, from size of legend box and of plot
        ymin, Imax = plt.ylim()
        for (pol, style) in [(policy, linestyle),
                             #(rempolicy, "--")
                             ]:
            X = np.empty((N, N))
            Y = np.empty((N, N))
            Z = np.empty((N, N))
            for S in range(N):
                for I in range(N):
                    X[(S,I)] = S
                    Y[(S,I)] = I
                    if (S,I) in pol:
                        Z[(S,I)] = 1.0
                        Imax = max(I, Imax)
                    else:
                        Z[(S,I)] = 0.0
            plt.contour(X, Y, Z, levels=[0.5], colors=[color], linestyles=[style])
            plt.plot([], [], color=color, linestyle=linestyle, label=label) # a dummy to make the legend
        plt.ylim(ymin=0, ymax=Imax)
    else:
        Slimit = sirdefs.get_Slimit(N, u1)
        Slist = []
        Imax = []
        for S in range(1, Slimit+1):
            Slist.append(S)
            Imax.append(0)
        for (S,I) in policy:
            if S>0:
                try:
                    Imax[S-1] = max(Imax[S-1], I+1) # maybe add 0.5 to indicate crossover is between two integers?
                except IndexError:
                    raise Exception, "IndexError on (%d,%d)" % (S,I)
        not_in_policy = []
        for S in Slist:
            for I in range(1, Imax[S-1]):
                if (S,I) not in policy | rempolicy:
                    not_in_policy.append((S,I))
        if len(not_in_policy):
            sys.stdout.write("States which aren't in policy, but are below a state which is (which is unexpected):\n" + str(not_in_policy) + "\n")
        line, = plt.plot(Slist, Imax, color=color, linestyle=linestyle, label=label)
        if linestyle=="--":
            # using this example to customise linestyles:
            # http://matplotlib.org/examples/lines_bars_and_markers/line_demo_dash_control.html
            line.set_dashes([20, 5])
        elif linestyle==":":
            line.set_dashes([5, 5])
        ymax = max(Imax)
        #plt.ylim(ymin=0, ymax=max(ymax, plt.ylim()[1]))

def check(N, B1, B2, y, u1, u2, u3, I0, printit=0, plotit=0, eval=0, evaltoggle=[]):

    #print "B=%g, (Sd,Id) = (%d,%d), Iupper2=%d" % (B1, Sd,Id, Iupper2)
    #sys.stdout.write("N=%d, B1=%g, y=%g, u1=%g, u2=%g, u3=%g, (Sd,Id)=(%d,%d), B2=%g\n"
    #                     % (N, B1, y, u1, u2, u3, Sd, Id, B2))

    cond = "%d_%g_%g_%g_%g_%g_%g" % (N, B1, B2, y, u1, u2, u3)

    if u1 > 0.0:
        Nfixed = 0
    else:
        Nfixed = 1

    S0 = N - I0
    (Sd1, Id1) = sirdefs.stable(N, B1, y, u1, u2, u3)
    (Sd, Id) = sirdefs.stable(N, B2, y, u1, u2, u3)
    (Se1, Ie1) = sirdefs.stable_float(N, B1, y, u1, u2, u3)
    (Se, Ie) = sirdefs.stable_float(N, B2, y, u1, u2, u3)
    Iupper = Id
    Iupper2 = sirdefs.get_Iupper2(N, y, u1, u2)
    Ipolicytop = sirdefs.get_Ipolicytop(N, y, u1, u2)

    policies = {}

    plotinit = 0
    if plotit:
        plt.clf()
        # http://tex.stackexchange.com/questions/109703/how-to-determine-the-font-being-used-by-a-latex-document
        # \showthe\font in the latex showed that OT1/cmr/m/n/12 (cmr = Computer Modern Roman) was the font
        plt.rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
        plt.rc('text', usetex=True)
        plotapprox(N, B1, B2, y, u1, u2, u3, Nfixed, color="g")
    for filename, color, linestyle, label, loop in [
        ("policyiter_"+cond+".jres", "b", "-", r"idealised scenario - optimal policy", 1), # lightgray
        ("linprogcrude_"+cond+".jres", "r", "-", r"realistic scenario - optimal policy", 2), # #A0A0A0, a bit lighter than gray
        #("../old/control/policyiter_"+cond+".jres", "b", ":", "unlimited switching (old)", 4),
        #("../old/control/linprogcrude_"+cond+".jres", "r", ":", "single switch (old)", 5),
        #("policyiterupper_"+cond+".jres", "c", "unlimited, higher $I_d$", 4),
        #("linprogcrudeupper_"+cond+".jres", "orange", "single switch, higher $I_d$", 5),
        #("policyiter3_"+cond+".jres", "g", "unlimited, crooked upper bound", 6),
        #("linprogj_"+cond+".jres", "r", "single switch (Clp solver)", 4),
        #("linprogj_"+cond+"mod.jres", "r", "Best if single switch (LP)", 4),
        ]:
        if os.path.exists(filename):
            print ""
            policy = getpolicy.getpolicy(filename)
            fpolicy = frozenset(policy)
            if fpolicy in policies:
                print filename + " identical policy to " + policies[fpolicy]
            else:
                print filename + " policy (len %d) not seen before, adding to policies" % len(policy)
                sys.stdout.write("(Sd1,Id1)=(%d,%d), (Sd,Id)=(%d,%d)\n" % (Sd1, Id1, Sd, Id))
                policies[fpolicy] = filename
                if printit:
                    printpolicy(policy, filename)
                if eval:
                    for sticky in [1]:
                        for mode in [0]:
                            if mode==0:
                                thisu1 = u1
                                thisu2 = u2
                                thisu3 = u3
                            else:
                                if Nfixed: # switch to unfixed
                                    thisu1 = u2
                                    thisu2 = 0.0
                                    thisu3 = 0.0
                                else: # switch to fixed
                                    thisu1 = 0.0
                                    thisu2 = u1
                                    thisu3 = u1
                            thispolicy = set(policy)
                            policyname = filename
                            for SI in evaltoggle:
                                if SI in thispolicy:
                                    thispolicy.remove(SI)
                                    policyname = policyname + "-" + str(SI)
                                else:
                                    thispolicy.add(SI)
                                    policyname = policyname + "+" + str(SI)
                            p0 =  markovfull.markovfull("",
                                                        N=N, BN=B1, y=y, u1=thisu1, u2=thisu2, u3=thisu3, I0=I0,
                                                        Sd=Sd, Id=Iupper, Iupper2=Iupper2,
                                                        phase_method=1, timeit=0, verbose=0, S0=N-I0,
                                                        switchset=thispolicy, stickyswitch=sticky,
                                                        betamod=B2/float(B1), rem=1e-6)
                            print ("markovfull, policy from %s (len %d), sticky=%d, Nfixed=%d, (S0,I0)=(%d,%d), p0= %s"
                                   % (policyname, len(thispolicy), sticky, (Nfixed + mode) % 2, S0, I0, repr(p0)))
            if plotit:
                if not plotinit:
                    plotinit = 1
                    if 0:
                        sirplot.sirplot(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0,
                                        clf=0, normalise=0, det=1, randruns=0,
                                        xplot="s",
                                        yplot="i",
                                        xlabel="S",
                                        ylabel="I",
                                        ylabel_rotation="horizontal",
                                        #title="5 simulation runs + deterministic",
                                        tmax=30,
                                        plotargs={"i" : {"color" : "k",
                                                         "linewidth" : 3, "linestyle" : "dashed",
                                                         "label" : "deterministic path"}
                                                  },
                                        )
                    tup = time.localtime()
                    if u2==0 and u3==0:
                        plt.title(r"Use (or start using) $\beta^{(2)}$ under the line", fontsize="x-large")
                        #+ "\n"
                        #              + r"($N=%d$, $\beta^{(1)}=%g$, $\beta^{(2)}=%g$, $\gamma=%g$, $\mu=%g$)"
                        #              % (N, B1, B2, y, u1))
                    else:
                        plt.title(r"Use $\beta^{(2)}$ on or below the line." # + "%s\n" % (time.strftime("%H:%M", tup))
                                  + r"$N=%d$, $\beta^{(1)}=%g$, $\beta^{(2)}=%g$, $\gamma=%g$, $\mu_1=%g$, $\mu_2=%g$, $\mu_3=%g$"
                                  % (N, B1, B2, y, u1, u2, u3), fontsize="large")
                    #if N==300:
                    plt.xlabel("S", fontsize="large")
                    plt.ylabel("I", rotation="horizontal", fontsize="large")
                plotpolicy(policy, set(), N, B1, y, u1, u2, u3, color, linestyle, label, Ipolicytop)
        else:
            print filename + " not found"
    if plotit:
        plt.plot([Sd1, Sd, Sd, N], [Id1, Id1, Id, Id], color="k", linestyle="--", linewidth=3,
                 label="$U$, the upper absorbing boundary")
        plt.scatter(Sd1, Id1, color="k", s=40)
        plt.scatter(Sd,  Id,  color="k", s=40)
        plt.text(Sd1, Id1+1, r"$\left(S_d^{(1)},I_d^{(1)}\right)$", color="k", fontsize="large", va="bottom")
        plt.text(Sd,  Id-1,  r"$\left(S_d^{(2)},I_d^{(2)}\right)$", color="k", fontsize="large", va="top")
        plt.legend(loc="lower left", handlelength=4, fontsize="large") # not supported at home
        #plt.scatter(Se1, Ie1, color="k")
        #plt.scatter(Se, Ie, color="k")
        plt.xlim(xmin=0, xmax=N)
        plt.ylim(ymin=0) #, ymax=2*Id1+1)
        locs, labels = plt.xticks()
        plt.xticks(locs, fontsize="large")
        locs, labels = plt.yticks()
        plt.yticks(locs, fontsize="large")
        #plt.savefig("policy_" + cond + ".png")
        plt.savefig("policy_" + cond + ".pdf")

