# 2018 update
# these were added because "iceil" and "itrunc" has been superseded in julia

function iceil(x)
    return ceil(Int, x)
end

function itrunc(x)
    return trunc(Int, x)
end

#######################

function sir_stable_float(N, BN, y, u1, u2, u3)
    B = BN / N
    Se = (y + u1 + u3) / B
    Ie = (N - Se) * (u1 + u2) / (y + u1 + u2)
    return (Se, Ie)
end

# Note backward incompatability in rare cases, as with sir_stable.py
# e,g, N=100, BN=3, y=1, u=ui=0.2 which gives Ie  = 10.000000000000002 = 10 + 2e-15.
# rev 1.1: Id = 11 (machine code correct)
# rev 1.2: Id = 10 (mathematically and intuitively correct)

# we round up, then test for >= in the code.
# e.g. escape when I>=Id and S>=Sd
# which is equivalent to I>Ie and S>Se
function sir_stable(N, BN, y, u1, u2, u3)
    (Se, Ie) = sir_stable_float(N, BN, y, u1, u2, u3)
    # Note that if julia (and python) rounding means Ie = 10.000000000000002 = 10 + 2e-15,
    # (which we actually get for N=100, BN=3, y=1, u=ui=0.2)
    # then Id=10 and so Id > Ie. THIS IS WHAT WE WANT.
    # i.e. Id - Ie should be between -1e-12 and 1-1e-12
    Sd = iceil(Se - 1e-12)
    Id = iceil(Ie - 1e-12)
    return (Sd, Id)
end

# right: N*u1 + (N-S-I)*u2
# left: S*u1
# down: (y+u1)*I
# balanced:
# (y+u1)*I = (N-S)*u1 + (N-S-I)*u2
# => (y+u1+u2)*I = (N-S)*(u1+u2)
# => I = (N-S)*(u1+u2) / (y+u1+u2)
function get_Ipolicytop(N, y, u1, u2)
    # u1(N-S) = (y+u1)I
    # S=0: I = N * u1 / (y+u1)
    x = iceil(N*(u1+u2) / (y+u1+u2)) + 1
    return x
end

function get_Iupper2(N, y, u1, u2)
    x = get_Ipolicytop(N, y, u1, u2)
    return 2*x
end

function get_Slimit(N, u1)
    if u1 > 0.0
        #if N < 1000
        #    return N + 100
        #else
        return itrunc(N * 1.1 + 0.5)
    else
        return N-1 # since S=N would mean I=0
    end
end

function get_Ilimit(N, u1)
    if u1 > 0.0
        return itrunc(N * 1.1 + 0.5)
    else
        return N
    end
end

# return whether we hit upper boundary.
# Intention is that this same module is called from all code, like sir_stable

# Note that since indexing starts at S==0, row size is srowmax+1
# Sd1<=Sd, Id1>=Id
function Srowmax(Slimit, Ilimit,Se1, Ie1, Se, Ie, I)
    Id =  iceil(Ie  - 1e-12) # identical to code in sir_stable.jl
    Id1 = iceil(Ie1 - 1e-12) # identical to code in sir_stable.jl
    Sd  = iceil(Se  - 1e-12) # identical to code in sir_stable.jl
    Sd1 = iceil(Se1 - 1e-12) # identical to code in sir_stable.jl
    if I > Ilimit
        return -1
    elseif I > Id1
        return Slimit
    elseif I == Id1 # Do in this order to catch the case of Id1==Id
        return Sd1 - 1
    elseif I >= Id
        return Sd - 1
    else
        return Slimit
    end
end

function upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
    if S > Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I)
        return true
    else
        return false
    end
end
