import sys
import math

def stable_float(N, BN, y, u1, u2, u3):
    B = BN / float(N)
    Se = (y + u1 + u3) / float(B)
    Ie = (N - Se) * (u1 + u2) / float(y + u1 + u2)
    return (Se, Ie)

# Note backward incompatability in rare cases.
# e,g, N=100, BN=3, y=1, u=ui=0.2 which gives Ie  = 10.000000000000002 = 10 + 2e-15.
# rev 1.2: Id = 11 (machine code correct)
# rev 1.3: Id = 10 (mathematically and intuitively correct)

# we round up, then test for >= in the code.
# e.g. escape when I>=Id and S>=Sd
# which is equivalent to I>Ie and S>Se
def stable(N, BN, y, u1, u2, u3):
    (Se, Ie) = stable_float(N, BN, y, u1, u2, u3)
    # Note that if python rounding means Ie = 10.000000000000002 = 10 + 2e-15,
    # (which we actually get for N=100, BN=3, y=1, u=ui=0.2)
    # then Id=10 and so Id > Ie. THIS IS WHAT WE WANT.
    # i.e. Id - Ie should be between -1e-12 and 1-1e-12
    Sd = int(Se - 1e-12) + 1
    Id = int(Ie - 1e-12) + 1
    return (Sd, Id)

# right: N*u1 + (N-S-I)*u2
# left: S*u1
# down: (y+u1)*I
# balanced:
# (y+u1)*I = (N-S)*u1 + (N-S-I)*u2
# => (y+u1+u2)*I = (N-S)*(u1+u2)
# => I = (N-S)*(u1+u2) / (y+u1+u2)
def get_Ipolicytop(N, y, u1, u2):
    # u1(N-S) = (y+u1)I
    # S=0: I = N * u1 / (y+u1)
    x = int(N*(u1+u2) / (y+u1+u2) - 1e-12) + 2 # yes, +2. It's iceil+1 in the Julia
    return x

def get_Iupper2(N, y, u1, u2):
    return 2 * get_Ipolicytop(N, y, u1, u2)

def get_Slimit(N, u1):
    if u1 > 0.0:
        #if N < 1000:
        #    return N + 100
        #else:
        return int(N * 1.1 + 0.5)
    else:
        return N-1 # since S=N would mean I=0

def get_Ilimit(N, u1):
    if u1 > 0.0:
        return int(N * 1.1 + 0.5)
    else:
        return N
    
# return whether we hit upper boundary.
# Intention is that this same module is called from all code, like sir_stable
#cy-
def    Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I):
#cy+
#cpdef Srowmax(int Slimit, int Ilimit, double Se1, double Ie1, double Se, double Ie, int I):
    #cdef int Id, Id1, Sd, Sd1
#cy.
    Id  = int(Ie - 1e-12)  + 1 # identical to code in sir_stable.py
    Id1 = int(Ie1 - 1e-12) + 1 # identical to code in sir_stable.py
    Sd  = int(Se - 1e-12)  + 1 # identical to code in sir_stable.py
    Sd1 = int(Se1 - 1e-12) + 1 # identical to code in sir_stable.py
    if I > Ilimit:
        return -1
    elif I > Id1:
        return Slimit
    elif I == Id1: # Do in this order to catch the case of Id1==Id
        return Sd1 - 1
    elif I >= Id:
        return Sd - 1
    else:
        return Slimit

#cy-
def    upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I):
#cy+
#cpdef upperboundary(int Slimit, int Ilimit, double Se1, double Ie1, double Se, double Ie, int S, int I):
#cy.
    if S > Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I):
        return True
    else:
        return False

# This is clumsy, but ensures the same check is used as in Srowmax
# Note the inconsistency that (Srowmax,I) is in the region
# but (S, Irowmax) is outside (it's the upper absorbing bound)
#cy-
def    Icolmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, Iupper2, S):
#cy+
#cpdef Icolmax(int Slimit, int Ilimit, double Se1, double Ie1, double Se, double Ie, int Iupper2, int S):
    #cdef int I
#cy.
    for I in range(Iupper2+1):
        # find highest I which is not in the region
        if S > Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I):
            # i.e. if (S,I) is NOT in the region
            return I
    # region goes to >= Iupper2, so return Iupper2
    return Iupper2


# find max I in the policy, in this column
def policy_max_this_column(S, N, Id, Sx, Ix, switchset, switch_ABC):
    if len(switchset):
        Imax = 0
        for (SS, I) in switchset:
            if SS==S:
                Imax = max(Imax, I)
        return Imax
    elif len(switch_ABC):
        if len(switch_ABC)==3:
            # standard form: AS + BI < C (for S < Sd), where A B and C may be floats
            # to be sure that rounding is always handled the same way
            stdA = switch_ABC[0]
            stdB = switch_ABC[1]
            stdC = switch_ABC[2]
            for I in range(N):
                if stdA*S + stdB*I < stdC:
                    Imax = I
                else:
                    return max(Imax, Id)
            else:
                    raise Exception, "no limit to Imax for S=%d , switch_ABC=%s and N=%d" % (S, switch_ABC, N)
        else:
            raise Exception, "switch_ABC must be length 3 or 0"
    elif len(Sx):
        if len(Sx) != len(Ix):
            raise Exception, "List length mismatch, len(Sx)=%d, len(Ix)=%d" % (len(Sx), len(Ix))
        for j in range(1, len(Sx)):
            if Sx[j] <= Sx[j-1]:
                raise Exception, "Sx values must be increasing: %g, %g; %s" % (Sx[j-1], Sx[j], Sx)
            if S==Sx[j-1]:
                return Ix[j-1]
            elif S==Sx[j]:
                return Ix[j]
            elif S < Sx[j]:
                SIslope = (Ix[j] - Ix[j-1]) / float(Sx[j] - Sx[j-1])
                Imax = int(Ix[j-1] + (S-Sx[j-1])*SIslope)
                return Imax
        else:
            raise Exception, "no range found for S=%d, Sx=%s" % (S, Sx)
    else:
        return 0
