import sys
import random
import matplotlib.pyplot as plt

import sir_stable

def randomstep(S, I, N, BN, y, u1, u2, u3):
    R = N - (S+I)
    Irenew = 0.0 + I * u3
    Rrenew = Irenew + R * u2 + N * u1
    Infect = Rrenew + I * BN * S/float(N)
    Recover = Infect + I * (y + u1)
    Sdeath = Recover + S * u1
    #sys.stdout.write("%d, %d: %.4f  %.4f  %.4f  %.4f  %.4f\n" % (S, I, Irenew, Rrenew, Infect, Recover, Sdeath))
    dt = random.expovariate(Sdeath)
    x = random.random() * Sdeath
    if x < Irenew:
        tup = (dt, 1, -1) # Irenew:  S++, I--
    elif x > Irenew and x < Rrenew:
        tup = (dt, 1, 0)  # Rrenew:  S++
    elif x > Rrenew and x < Infect:
        tup = (dt, -1, 1) # Infect:  S--, I++
    elif x > Infect and x < Recover:
        tup = (dt, 0, -1) # Recover: I--
    elif x > Recover:
        tup = (dt, -1, 0) # Sdeath:  S--
    else:
        # the unlikely case of an exact ==, do nothing on these to avoid bias
        tup = (0, 0, 0)
    return tup

def cplot(xlist, ylist, xname, yname, log, plotargs, defaultcolors, runlabel,
          xoffset, yoffset, plot):
    if xoffset or yoffset:
        for i in range(len(xlist)):
            xlist[i] = xlist[i] - xoffset
            ylist[i] = ylist[i] - yoffset
    if plotargs and yname in plotargs:
        kwargs = plotargs[yname]
    else:
        kwargs = {}
    if "label" not in kwargs:
        kwargs["label"] = yname
    # some default colors
    if defaultcolors and "color" not in kwargs:
        if yname[0]=="s":
            kwargs["color"] = "b"
        elif yname[0]=="i":
            kwargs["color"] = "r"
        elif yname[0]=="r":
            kwargs["color"] = "g"
        elif yname=="N":
            kwargs["color"] = "m"
    # some default linestyles
    if defaultcolors and "linestyle" not in kwargs:
        if yname.count("_stable"):
            kwargs["linestyle"] = "--"
    if plot:
        plt.plot(xlist, ylist, **kwargs)
    if log:
        log.write(runlabel + " " + xname + " (len=" + str(len(xlist)) + ") " + str(xlist) + "\n")
        log.write(runlabel + " " + yname + " (len=" + str(len(ylist)) + ") " + str(ylist) + "\n")

def sirplot(N=None,      # population size
            BN=None,     # BN/N = infection factor beta
            y=None,      # recovery rate gamma
            u1=0,        # birth and death rate (unfixed pop size model)
            u2=0,        # R -> S (loss of immunity) rate u in fixed size model
            u3=0,        # I -> S rate (death/rebirth of infectives) in fixed size model. 0 in SIRS model.
            I0=1,        # initial number of infectious
            S0=-1,       # initial number susceptible. If S0<0, use N-I0-R0
            R0=0,        # initial number recovered. Ignored if S0>=0
            tmax=10.0,   # time (x dimension) limit. Not the same as xmax. 0 or <0 means no limit
            infmax=0,    # max number of infection events. 0 or <0 means no limit.
            evmax=0,     # max number of events. 0 or <0 means no limit.
            Imax=-1,      # exit at this I value. <0 means no Imax
            Imin=-1,      # exit at this I value. <0 means no Imin
            Smin=-1,      # exit at this S value. <0 means no Smin
            ymin=0,      # y min on plot
            ymax=None,   # y max on plot. None means no maximum
            xmin=0,      # x min on plot
            xmax=None,   # x max on plot. None means no maximum
            xoffset=0,   # plot x-xoffset everywhere
            yoffset=0,   # plot y-yoffset everywhere
            randomcolors=0, # turn off any sirplot.py defaults, and let matplotlib allocate colors instead
            plotargs=None, # dictionary, each key is a yname and each value is a kwargs dict
            #                kwargs keys may include "label", "color", "linestyle",
            #                for a full list see matplotlib.pyplot.plot kwargs
            legend=None, # do a little legend in the top corner. Needs enhancing
            normalise=1, # scale variables to a fraction of 1
            det=1,       # do deterministic plot
            tstep=0.001, # time step for deterministic sim
            randruns=0,  # number of random runs
            readlogfile="", # log file to read, instead of re-running random sims
            title=None,  # title at top of plot
            xlabel=None, # label to put on x axis (otherwise use xplot)
            ylabel=None, # label to put on y axis
            ylabel_rotation="vertical",
            numbers_on_axes=True, # if False, no numbers on axes
            yplot="i",    # comma separated list of what to plot in y direction. List for subplots
            # yplot possibilities are "i", "r", "s", "i_stable", "s_stable", "r_stable"
            xplot="t",   # x axis plot, default is "t" (time)
            clf=1,       # clear plot. Setting 0 useful if doing multiple calls of sirplot
            adjust_left=None, # make extra room at left. Sometimes needed for larger axis font
            adjust_right=None, # make extra room at right. Sometimes needed for larger axis font
            adjust_bottom=None, # make extra room at bottom. Sometimes needed for larger axis font
            adjust_top=None, # make extra room at top. Seems no help anyway
            aspect=None, # specify an aspect ration
            figsize=None, # 2-tuple specifying size in inches. If not specified, get (8,6)
            save="",     # file name to save plot.  Should end in ".png".
            logfile="",  # log file to save (might be huge)
            summaryfile="", # summary file to save.
            plot=1,      # set 0 to go through the motions but not do the pyplot.plot command
            ret=None):   # string which gets eval'ed to give quantity to return

    if N==None or BN==None or y==None:
        raise Exception, "N, BN and y must be set"

    B = BN / float(N)

    list = {} # dictionary of plot lists

    yplots = []
    subplots = []
    if yplot:
        if type(yplot) == str: # str is same as types.StringType
            items = yplot.split(",")
            yplots.extend(items)
            # subplots stays as []
        else:
            subplotcount = len(yplot)
            nroot = 100 * subplotcount + 11
            for i in range(len(yplot)):
                item = yplot[i]
                items = item.split(",")
                yplots.extend(items)
                # first item gives subplot param (e.g. 211 and 212 if len(yplot)==2)
                subplots.append([nroot+i] + items)

    if logfile:
        log = open(logfile, "w")
    else:
        log = None

    if summaryfile:
        summary = open(summaryfile, "w")
        summary.write("run\ttendtime\tinfections\tevents\n")
    else:
        summary = None

    if normalise:
        randscale = 1 / float(N)
        detscale = randscale #1
    else:
        randscale = 1
        detscale = randscale #N

    # Turn off defaults if more than 1 rand run,
    # because then you get lots of plots the same color
    defaultcolors = (not randomcolors) and randruns <= 1

    if clf:
        plt.clf()

    if ymax:
        # note this picks up the ymin default (0)
        plt.ylim(ymin=ymin, ymax=ymax)
    elif ymin: # can't imagine setting ymin only, but I guess it's possible
        plt.ylim(ymin=ymin)

    if xmax:
        # note this picks up the xmin default (0)
        plt.xlim(xmin=xmin, xmax=xmax)
    elif xmin: # can't imagine setting xmin only, but I guess it's possible
        plt.xlim(xmin=xmin)

    if xlabel:
        plt.xlabel(xlabel)
    else:
        plt.xlabel(xplot)

    if ylabel:
        plt.ylabel(ylabel, rotation=ylabel_rotation)

    if title:
        plt.title(title)

    # Srate = -B*S*I + I*u3 + (N-S-I)*u2 + N*u1 - S*u1
    # Irate = B*S*I - y*I - I*u1 - I*u3
    # Irate = 0
    #  -> B*S = y + u1 + u3
    #  -> S := Se = (y + u1 + u3)/B
    #             = N(y+u1+u3)/(BN)
    # 
    # Srate = 0 and S = Se = (y + u1 + u3)/B
    #  -> -B*Se*I + I*u3 + (N-Se-I)*u2 + (N-Se)*u1 = 0
    #  -> -B*((y+u1+u3)/B)*I + I*u3 - I*u2 + (N-Se)*(u1+u2) = 0
    #  -> -(y+u1+u3)*I + I*u3 - I*u2 + (N-Se)*(u1+u2) = 0
    #  -> -(y+u1+u2)*I + (N-Se)*(u1+u2) = 0
    #  -> I = (N-Se)*(u1+u2) / (y+u1+u2)
    #
    # always a flat or vertical line
    stablescale = randscale
    (Se, Ie) = sir_stable.stable_float(N, BN, y, u1, u2, u3)
    list["s_stable"] = [stablescale * Se]
    list["i_stable"] = [stablescale * Ie]
    list["r_stable"] = [stablescale * (N - Se - Ie)]

    if readlogfile:
        ip = open(readlogfile, "r")

    if S0 < 0:
        S0 = N - I0 - R0
    else:
        if R0 > 0 and R0 != N-I0-S0:
            sys.stderr.write("Warning: S0/R0 overspecified, ignoring R0=%d, using I0=%d and S0=%d which gives R0=%d\n"
                                 % (R0, I0, S0, N-I0-S0))

    for run in range(randruns):
        t = 0.0
        events = 0
        I = I0
        S = S0
        infections = I0
        list["t"] = []
        list["events"] = []
        list["infections"] = []
        list["s"] = []
        list["i"] = []
        list["r"] = []
        if readlogfile: # read old random sim
            # skip to first non-empty line (or EOF)
            while 1:
                line = ip.readline()
                words = line.split()
                if line=="" or len(words):
                    break
            if line=="":
                break # EOF
            while 1: # read data until empty line or EOF
                bracket = line.find("[")
                if words[1] in list and bracket > 0:
                    list[words[1]] = eval(line[bracket:])
                elif words[1] not in list:
                    raise Exception, "bad list key: " + str(words[1]) + "\nlist = " + str(list)
                elif len(words) >= 3:
                    raise Exception, "couldn't find list for line beginning: " + words[0] + " " + words[1] + " " + words[2]
                else:
                    raise Exception, "couldn't find list for line: " + line
                # get next line
                line = ip.readline()
                words = line.split()
                if len(words)==0:
                    break
        else: # do a random sim
            while ((t < tmax or tmax<=0) and (infections < infmax or infmax<=0) and (events < evmax or evmax<=0)
                   and (I < Imax or Imax < 0) and (I > Imin or Imin < 0)
                   and (S > Smin or Smin < 0)):
                R = N - S - I
                list["t"].append(t)
                list["events"].append(events)
                list["infections"].append(infections)
                list["s"].append(S*randscale)
                list["i"].append(I*randscale)
                list["r"].append(R*randscale)
                if I==0:
                    break
                (dt, dS, dI) = randomstep(S, I, N, BN, y, u1, u2, u3)
                t += dt
                S += dS
                I += dI
                events += 1
                if dI==1:
                    infections += 1
        if len(subplots)<=1:
            for yname in ("s", "i", "r"):
                if yname in yplots:
                    cplot(list[xplot], list[yname], xplot, yname, log, plotargs, defaultcolors, "rand"+str(run), xoffset, yoffset, plot)
                for yname in ("s_stable", "i_stable", "r_stable"):
                    if yname in yplots:
                        cplot(list[xplot], len(list[xplot])*list[yname], xplot, yname, log, plotargs, defaultcolors, "rand"+str(run), xoffset, yoffset, plot)
            if log:
                log.write("\n") # only blank line after each random run (not after each param pair)
        else:
            for sp in subplots:
                plt.subplot(sp[0])
                for yname in ("s", "i", "r"):
                    if yname in sp:
                        cplot(list[xplot], list[yname], xplot, yname, log, plotargs, defaultcolors, "rand"+str(run), xoffset, yoffset, plot)
                    for yname in ("s_stable", "i_stable", "r_stable"):
                        if yname in sp:
                            cplot(list[xplot], len(list[xplot])*list[yname], xplot, yname, log, plotargs, defaultcolors, "rand"+str(run), xoffset, yoffset, plot)
            if log:
                log.write("\n") # only blank line after each random run (not after each param pair)
        if run==0 and legend:
            if legend==1:
                plt.legend()
            else:
                plt.legend(legend)
        if summary:
            summary.write("%d\t%g\t%d\t%d\n" % (run, t, infections, events))

    if readlogfile:
        ip.close()

    if det:
        # on 22-Jun-2016, moved this line from out of the loop: scaling shouldn't change with N
        if normalise:
            detscale = 1 / float(N)
        else:
            detscale = 1
        i = I0
        s = S0
        r = N - (s+i)
        t = 0.0
        events = 0.0
        infections = 0.0
        smin = Smin / float(N)
        imin = Imin / float(N)
        imax = Imax / float(N)
        list["t"] = []
        list["events"] = []
        list["infections"] = []
        list["recovers"] = []
        list["rrenews"] = []
        list["s"] = []
        list["i"] = []
        list["r"] = []
        maxs = {} # list of (xplot,yplot) tuple pairs of maxima
        mins = {} # list of (xplot,yplot) tuple pairs of minima
        for yname in yplots:
            maxs[yname] = []
            mins[yname] = []
        while ((t < tmax or tmax<=0) and (infections < infmax or infmax<=0) and (events < evmax or evmax<=0)
               and (i < Imax or Imax < 0) and (i > Imin or Imin < 0)
               and (s > Smin or Smin < 0)):
            list["t"].append(t)
            list["s"].append(s*detscale)
            list["i"].append(i*detscale)
            list["r"].append(r*detscale)
            if len(list["events"])==0:
                list["events"].append(0.0)
                list["infections"].append(0.0)
                list["recovers"].append(0.0)
                list["rrenews"].append(0.0)
            else:
                list["events"].append(list["events"][-1] + totalrate * detscale * tstep)
                list["infections"].append(list["infections"][-1] + infrate * detscale * tstep)
                list["recovers"].append(list["recovers"][-1] + recrate * detscale * tstep)
                list["rrenews"].append(list["rrenews"][-1] + rrenewrate * detscale * tstep)
            # s_balanced is fixed if N does not change. So remove that code.
            for yname in yplots:
                if yname in list and len(list[yname]) >= 3:
                    if list[yname][-1] > list[yname][-2] and list[yname][-2] <= list[yname][-3]:
                        mins[yname].append( (list[xplot][-2], list[yname][-2]))
                    elif list[yname][-1] < list[yname][-2] and list[yname][-2] >= list[yname][-3]:
                        maxs[yname].append( (list[xplot][-2], list[yname][-2]))
            events     = list["events"][-1]
            infections = list["infections"][-1]
            infrate = B*s*i # need this so we can count infections
            recrate = y*i
            irenewrate = u3*i
            rrenewrate = u2*r
            birthrate = u1*N
            sdeathrate = u1 * s
            ideathrate = u1 * i
            rdeathrate = u1 * r
            totalrate = (infrate + recrate + irenewrate + rrenewrate
                         + birthrate + sdeathrate + ideathrate + rdeathrate)
            srate = -infrate + irenewrate + rrenewrate + birthrate - sdeathrate
            irate = infrate - recrate - irenewrate - ideathrate
            rrate = recrate - rdeathrate - rrenewrate
            s = s + srate*tstep
            i = i + irate*tstep
            r = r + rrate*tstep
            t += tstep
        if len(subplots)<=1:
            for yname in ("s", "i", "r", "infections", "recovers", "rrenews"):
                if yname in yplots:
                    cplot(list[xplot], list[yname], xplot, yname, log, plotargs, defaultcolors, "det", xoffset, yoffset, plot)
            for yname in ("s_stable", "i_stable", "r_stable"):
                if randruns==0 and yname in yplots:
                    cplot(list[xplot], len(list[xplot])*list[yname], xplot, yname, log, plotargs, defaultcolors, "det", xoffset, yoffset, plot)
        else:
            for sp in subplots:
                plt.subplot(sp[0])
                for yname in ("s", "i", "r", "infections", "recovers", "rrenews"):
                    if yname in sp:
                        cplot(list[xplot], list[yname], xplot, yname, log, plotargs, defaultcolors, "det", xoffset, yoffset, plot)
                for yname in ("s_stable", "i_stable", "r_stable"):
                    if randruns==0 and yname in sp:
                        cplot(list[xplot], len(list[xplot])*list[yname], xplot, yname, log, plotargs, defaultcolors, "det", xoffset, yoffset, plot)
        if randruns==0 and legend:
            if legend==1:
                plt.legend()
            else:
                plt.legend(legend)

    #plt.gca().tight_layout() # crashes
    # From http://stackoverflow.com/questions/6774086/why-is-my-xlabel-cut-off-in-my-matplotlib-plot
    # works, but I'm a little concerned about putting a hard number there
    # gcf() = get current figure
    if adjust_left:
        plt.gcf().subplots_adjust(left=adjust_left)
    if adjust_right:
        plt.gcf().subplots_adjust(right=adjust_right)
    if adjust_bottom:
        plt.gcf().subplots_adjust(bottom=adjust_bottom)
    if adjust_top:
        plt.gcf().subplots_adjust(top=adjust_top)

    if figsize:
        if len(figsize)==2:
            plt.figure(figsize=figsize)
        else:
            raise Exception, "figsize needs to be a 2-tuple but instead figsize = " + str(figsize)

    if not numbers_on_axes:
        # don't display numbers (in fact, make them white)
        plt.tick_params(labelcolor="w")

    if save:
        plt.savefig(save)
    if logfile:
        log.close()
    if summaryfile:
        summary.close()
    if ret:
        return eval(ret)
        #return tuple(list["i"]), tuple(list["s"])
