import sys
import os
import time
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np

import sirdefs
import getpolicy
import sirdetpeaks
import sweeptroughfull
import markovfull
import sirplot

def make_approx_policy(N, y, u1, u2, u3, Sd, Id):
    policy = set()
    for I in range(1, N):
        for S in range(N):
            if S >= Sd and I < Id:
                policy.add((S,I))
            else:
                # Itop = top of uR = yI line
                # => uN - uS =(y+u)I when S=0
                # => uN = (y+u)I
                # => I = N u/(y+u)
                #if (y+u1)*I < u1*(N-S):
                if (y+u1)*I + u1*S < u1*N:
                    policy.add((S,I))
    return frozenset(policy)

def make_pwl_policy(N, Sx, Ix, verbose):
    if len(Sx) != len(Ix):
        raise Exception, "Warning, len(Sx)=%d != len(Ix)=%d\n%s\n%s" % (len(Sx), len(Ix), Sx, Ix)
    elif len(Sx) < 2:
        if len(Sx)==1:
            sys.stderr.write("Warning, cannot build a non-zero pwl policy when len(Sx)==1; Sx=%s, Ix=%s\n" % (Sx, Ix))
        return frozenset()
    else:            
        policy = set()
        for S in range(N):
            for j in range(len(Sx)):
                if S==Sx[j]:
                    Imax = int(Ix[j]) # round down
                    break
                elif j>0 or (S>Sx[j-1] and S<Sx[j]) or (S<Sx[j-1] and S>Sx[j]):
                    Imax = int(Ix[j-1] + (Ix[j] - Ix[j-1]) * (S - Sx[j-1]) / float(Sx[j] - Sx[j-1])) # round down
                    break
            else:
                raise Exception, "no match found for S=%d, Sx=%s, Ix=%s" % (S, Sx, Ix)
            for I in range(1, Imax+1): # up to and including rounded down value
                policy.add((S,I))
        if verbose:
            sys.stdout.write("make_pwl_policy: N=%d, Sx=%s, Ix=%s --> len(policy)=%d\n" % (N, Sx, Ix, len(policy)))
            if verbose >= 2:
                sys.stdout.write("make_pwl_policy: policy = %s\n" % policy)
        return frozenset(policy)

def printpolicy(policy, name):
    Ilist = []
    Slist = {}
    for (S,I) in policy:
        if I not in Ilist:
            Ilist.append(I)
            Slist[I] = [S]
        else:
            Slist[I].append(S)
    Ilist.sort()
    sys.stdout.write("policy %s:\n" % name)
    for I in Ilist:
        sys.stdout.write("I=%d:" % I)
        Slist[I].sort()
        for S in Slist[I]:
            sys.stdout.write(" %d" % S)
        sys.stdout.write(" (%d)\n" % len(Slist[I]))

# get the lowest left "hole" in the policy
def policy_mingap(policy, N):
    for I in range(1, N):
        for S in range(1, N):
            if (S,I) not in policy:
                return (S,I)
    raise Exception, "no hole found in policy length %d, N=%d\n" % (len(policy), N)

def evaluate(op,
             name="",
             N=0,  BN=0,  y=0,  u1=0,  u2=0, u3=0,
             I0=1, S0=-1,
             Sd1=-1, Id1=-1, Sd=-1,  Id=-1,  Iupper2=-1, Istart=-1,
             timeit=0, verbose=0,
             Sx=[], Ix=[], diffswitchcond="",
             switchset=set(), switch_ABC=[], stickyswitch=0,
             Ixb=-1, betamod=0.9, rem=1e-6,
             sweep=1,
             ):
    #sweep = N > 300
    if sweep:
        # note switch_ABC takes prcedence over Sx/Ix
        p0 = sweeptroughfull.sweeptroughfull(op,
                                             N, BN, y, u1, u2, u3, I0,
                                             #Sd, Id, Iupper2,
                                             #Istart=Istart,
                                             monitor_S=1, S0=S0,
                                             Sx=Sx, Ix=Ix, Ixb=Ixb, diffswitchcond=diffswitchcond,
                                             switchset=switchset, switch_ABC=switch_ABC,
                                             stickyswitch=stickyswitch,
                                             betamod=betamod,
                                             verbose=verbose)
        sys.stdout.write("sweeptroughfull, ")
    else:
        # make a policy from Sx and Ix
        if len(switchset)==0:
            if len(Sx)>=2:
                switchset = make_pwl_policy(N, Sx, Ix, verbose)
                if verbose>=2:
                    printpolicy(switchset, "deterministic")
        initphase = 0
        p0 =  markovfull.markovfull("",
                                    N=N, BN=BN, y=y, u1=u1, u2=u3, u3=u3, I0=I0,
                                    phase_method=1, timeit=timeit, verbose=verbose, S0=S0, initphase=initphase,
                                    Sx=Sx, Ix=Ix,
                                    switchset=switchset,
                                    stickyswitch=stickyswitch,
                                    betamod=betamod, rem=rem)
        sys.stdout.write("markovfull,      ")
    sys.stdout.write("policy %-60s, sticky=%d, (S0,I0)=(%d,%d), p0= %s\n"
                     % (name, stickyswitch, S0, I0, repr(p0)))
    if verbose:
        sys.stdout.write("\n")
    

def check(N, B1, B2, y, u1, u2, u3, I0, stickyswitch=0,
          do_unchanged=[], points=[], do_det_approx=0, do_best_policy=[], plotit=0, verbose=0, sweep=1):

    #print "B=%g, (Sd,Id) = (%d,%d), Iupper2=%d" % (B1, Sd,Id, Iupper2)
    #sys.stdout.write("N=%d, B1=%g, y=%g, u1=%g, u2=%g, u3=%g, (Sd,Id)=(%d,%d), B2=%g\n"
    #                     % (N, B1, y, u1, u2, u3, Sd, Id, B2))

    cond = "%d_%g_%g_%g_%g_%g_%g_%d" % (N, B1, B2, y, u1, u2, u3, I0)

    if u1 > 0.0:
        Nfixed = 0
    else:
        Nfixed = 1

    S0 = N - I0
    (Se1, Ie1) = sirdefs.stable_float(N, B1, y, u1, u2, u3)
    (Sd1, Id1) = sirdefs.stable(N, B1, y, u1, u2, u3)
    Iupper1 = Id1
    Iupper21 = sirdefs.get_Iupper2(N, y, u1, u2)
    Istart1 = Id1 #sirdefs.get_Ipolicytop(N, y, u1, u2)
    (Se, Ie) = sirdefs.stable_float(N, B2, y, u1, u2, u3)
    (Sd, Id) = sirdefs.stable(N, B2, y, u1, u2, u3)
    Iupper = Id
    Iupper2 = Iupper21
    Istart = Id #Istart1
    #Ipolicytop = sirdefs.get_Ipolicytop(N, y, u1, u2)

    if I0 < Id1:
        # Note these are the Id and Imax under B1, not B2. I'll call it Id1 to avoid confusion hopefully
        (Srisef, Irisef, trisef) = sirdetpeaks.detpoint(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                     point=1, forceint=0)
    else:
        (Srisef, Irisef) = (S0, I0)

    (Smaxf, Imaxf, tmaxf) = sirdetpeaks.detpoint(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                              point=2, forceint=0)
    (Smax, Imax, tmax) = sirdetpeaks.SIt_round(Smaxf, Imaxf, tmaxf, 1)

    sys.stdout.write("\n")
    sys.stdout.write("Irisef=%g, Imaxf=%g, (Sd1,Id1)=(%d,%d) (Sd,Id)=(%d,%d)\n"
                     % (Irisef, Imaxf, Sd1, Id1, Sd, Id))

    # start (just above Id1)
    (S1, I1, t1) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="I>=%g" % (Irisef), forceint=1)
    I1 += 1 #### a hack
    
    # half way up (from Id1)
    (S2, I2, t2) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="I>=%g" % ((Imaxf+Irisef)/2.0), forceint=1)
    # half way down (from Id1)
    (S3, I3, t3) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="S<%g and I<=%g" % (Se1, (Imaxf+Irisef)/2.0), forceint=1)
    # to ideal det point
    (S4, I4, t4) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="S<%g and %g*I < %g*(%d-S)" % (Se1, y+u1, u1, N), forceint=1)
    # Hitting Id (not Id1)
    (S5, I5, t5) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="S<%g and I<=%g or S>=%g and t>%g" % (Se1, Ie, Smaxf, tmaxf), forceint=1)
    # Hitting 1/4 Id (not Id1)
    (S6, I6, t6) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                         condstop="S<%g and I<=%g or S>=%g and t>%g" % (Se1, Ie/4.0, Smaxf, tmaxf), forceint=1)
    # Hitting S==Sd at min
    (S7, I7, t7) = sirdetpeaks.detpoint(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                              point=5, forceint=1)
    # to end point (next rise though Id1 if staying with Id1)
    (Send, Iend, tend) = sirdetpeaks.detpoint(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                              point=6, forceint=1)

    sys.stdout.write(cond + "\n")
    sys.stdout.write("(Sd1,Id1)=(%d,%d), (Se,Ie)=(%g,%g), (Sd,Id)=(%d,%d), Iupper2=%d, sticky=%d\n"
                     % (Sd1, Id1, Se, Ie, Sd, Id, Iupper2, stickyswitch))
    if verbose:
        sys.stdout.write("\n")

    for point in do_unchanged:
        if point==1:
            p0 = evaluate("", name="B=%g always (point F)" % (B1),
                          N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                          Sd1=Sd1, Id1=Id1, Sd=Sd, Id=Id, Iupper2=Iupper2, Istart=Istart1,
                          timeit=1, verbose=verbose, S0=S1,
                          switchset=set(),
                          stickyswitch=stickyswitch,
                          betamod=B2/float(B1), rem=1e-6, sweep=sweep)
        elif point==2:
            p0 = evaluate("", name="B=%g always (switch point A)" % (B2),
                          N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                          Sd1=Sd1, Id1=Id1, Sd=Sd, Id=Id, Iupper2=Iupper2, Istart=Istart1,
                          timeit=1, verbose=verbose, S0=S1,
                          Sx=[0, N], Ix=[N, N], # always switch
                          switchset=set(),
                          stickyswitch=stickyswitch,
                          betamod=B2/float(B1), rem=1e-6, sweep=sweep)
        else:
            raise Exception, "unknown unchanged point " + str(point)

    # different attempts
    for point in points:
        if point==2:
            [Sx, Ix, diffswitchcond, name] = [(0, Smax, Smax+1, N), (N, N, Id, Id), "", "start when I at peak (switch point B)"]
        elif point==3:
            [Sx, Ix, diffswitchcond, name] = [(0, S3, S3+1, N), (I3, I3, Id, Id), "", "start when I half down from peak to Id (switch point 3)"]
        elif point==5:
            [Sx, Ix, diffswitchcond, name] = [(0, Sd, Sd+1, N), (Id, Id, Id, Id), "", "start when I=Id on way down (switch point 5)"]
        elif point==6:
            [Sx, Ix, diffswitchcond, name] = [(0, Sd, Sd+1, N), ((1*Id)//4, (1*Id)//4, Id, Id), "", "start when I=Id/4 on way down (switch point D)"]
        elif point==7:
            [Sx, Ix, diffswitchcond, name] = [(0, Sd, Sd+1, N), (0, 0, Id, Id), "", "start when S=Sd at minimumum (switch point E)"]
        else:
            raise Exception, "unknown point specified: " + str(point)
        p0 =  evaluate("", name=name, #"from pwl [%s, %s]" % (Sx, Ix),
                       N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                       Sd=Sd, Id=Iupper, Iupper2=Iupper2, Istart=Istart,
                       timeit=1, verbose=verbose, S0=S1,
                       Sx=Sx, Ix=Ix, Ixb=0, diffswitchcond=diffswitchcond,
                       stickyswitch=stickyswitch,
                       betamod=B2/float(B1), rem=1e-6, sweep=sweep)

    # deterministic approximation
    if do_det_approx:
        Sx = [0, Se, Sd, N] # use by markovfull
        Ix = [N * u1 / float(y+u1), Ie, Id, Id]
        switch_ABC = [u1, y+u1, u1*N] # takes precedence over Sx/Ix for sweeptrough
        #printpolicy(policy, "deterministic")
        p0 =  evaluate("",
                       #name="from deterministic (S>=%d or %g*S + %g*I < %g) (switch point C)" % (Sd, switch_ABC[0], switch_ABC[1], switch_ABC[2]),
                       name="from deterministic (switch point C)",
                       N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                       Sd=Sd, Id=Iupper, Iupper2=Iupper2, Istart=Istart,
                       timeit=1, verbose=verbose, S0=S1,
                       Sx=Sx, Ix=Ix, Ixb=0,
                       switch_ABC=switch_ABC,
                       stickyswitch=stickyswitch,
                       betamod=B2/float(B1), rem=1e-6, sweep=sweep)

    if do_best_policy:
        policies = {}
        for loop, filename, color, label in [(0, "policyiter_"+cond+".jres", "r", "Best if infinite switching (MDP)"),
                                             #("linprogj_"+cond+".jres", "b", "Best if single switch (LP)"),
                                             (1, "linprogcrude_"+cond+".jres", "b", "Best if single switch (crude LP)"),
                                             ]:
            if os.path.exists(filename):
                policy = getpolicy.getpolicy(filename)
                fpolicy = frozenset(policy)
                if loop not in do_best_policy:
                    pass
                elif fpolicy in policies:
                    sys.stdout.write(filename + " identical policy to " + policies[fpolicy] + "\n")
                    if verbose:
                        sys.stdout.write("\n")
                else:
                    #print filename + " policy (len %d) not seen before, adding to policies" % len(policy)
                    policies[fpolicy] = filename
                    if verbose>=3:
                        printpolicy(fpolicy, filename)
                    p0 =  evaluate("", name="%s (len %d)" % (filename, len(policy)),
                                   N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                                   Sd=Sd, Id=Iupper, Iupper2=Iupper2, Istart=Istart,
                                   timeit=1, verbose=verbose, S0=S1,
                                   switchset=policy,
                                   stickyswitch=stickyswitch,
                                   betamod=B2/float(B1), rem=1e-6, sweep=sweep)
                    # check the various offsets if 2 (policyiter) or 3 (LP) in do_best_policy
                    if (loop+2) in do_best_policy:
                        if filename[:7]=="policyi" and stickyswitch==0 or filename[:7]=="linprog" and stickyswitch==1:
                            (Sgap,Igap) = policy_mingap(policy, N)
                            (Sgap,Igap) = (21,17)
                            policy.add((Sgap,Igap))
                            p0 =  evaluate("", name="%s + (%d,%d) (len %d)" % (filename, Sgap, Igap, len(policy)),
                                           N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                                           Sd=Sd, Id=Iupper, Iupper2=Iupper2, Istart=Istart,
                                           timeit=1, verbose=verbose, S0=S1,
                                           switchset=policy,
                                           stickyswitch=stickyswitch,
                                           betamod=B2/float(B1), rem=1e-6, sweep=sweep)
                            policy.remove((Sgap,Igap))
                            if Sgap > 1:
                                policy.remove((Sgap-1,Igap))
                                p0 =  evaluate("", name="%s - (%d,%d) (len %d)" % (filename, Sgap-1, Igap, len(policy)),
                                               N=N, BN=B1, y=y, u1=u1, u2=u3, u3=u3, I0=I1,
                                               Sd=Sd, Id=Iupper, Iupper2=Iupper2, Istart=Istart,
                                               timeit=1, verbose=verbose, S0=S1,
                                               switchset=policy,
                                               stickyswitch=stickyswitch,
                                               betamod=B2/float(B1), rem=1e-6, sweep=sweep)
                                policy.add((Sgap-1,Igap))
            else:
                sys.stdout.write(filename + " not found\n")
                
    if plotit:
        colors = list(cm.jet(np.linspace(0, 1, 9)))
        plt.clf()
        plt.figure(figsize=(8,3.5))
        xaxis = "s"
        Slist, tlist = sirplot.sirplot(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                       clf=0, det=1, normalise=0, tmax=tend,
                                       yplot="i", xplot=xaxis,
                                       plotargs={"i" : {"color" : "k",
                                                        #"label" : r"$\beta^{(1)}$ always",
                                                        "label" : "0",
                                                        "linestyle" : "--",
                                                        }},
                                       ret=('(list["s"],list["t"])'),
                                       )
        color = 1
        for (SS, II) in [(S1,I1),
                         (Smax,Imax), (S3,I3), (S4,I4), (S5,I5), (S6,I6),
                         (S7,I7),
                         ]:
            sirplot.sirplot(N=N, BN=B2, y=y, u1=u1, u2=u2, u3=u3, I0=II, S0=SS,
                            clf=0, det=1, normalise=0, tmax=tend,
                            yplot="i", xplot=xaxis,
                            plotargs={"i" : {"color" : colors[color],
                                             #"label" : r"$\beta^{(2)}$ from point %d" % color,
                                             "label" : "%d" % color,
                                             }},
                            )
            color += 1
        plt.legend()
        plt.yscale("log")
        plt.savefig("switchpathslog.png")
        plt.savefig("switchpathslog.pdf")
        plt.yscale("linear")
        plt.ylim(ymin=0, ymax=30)
        plt.savefig("switchpaths.png")
        plt.savefig("switchpaths.pdf")

        for xaxis in ["s", "t"]:
            plt.clf()
            plt.figure(figsize=(8,4.5))
            sirplot.sirplot(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                            clf=0, det=1, normalise=0, tmax=tend,
                            yplot="i", xplot=xaxis,
                            plotargs={"i" : {"color" : "r",
                                             }},
                            )
            
            if xaxis=="s":
                plt.text(S1, I1, "$1$", color="k")
                plt.text(Smax, Imax, "$2$", color="k")
                plt.text(S3, I3, "$3$", color="k")
                plt.text(S4, I4, "$4$", color="k")
                plt.text(S5, I5, "$5$", color="k")
                plt.text(S6, I6, "$6$", color="k")
                plt.text(S7, I7, "$7$", color="k")
                Ilist = []
                Idlist = []
                for S in Slist:
                    Ilist.append(u1*(N-S) / float(y+u1))
                    Idlist.append(Id)
                #plt.plot(Slist, Ilist, color="g")
                plt.plot(Slist, Idlist, color="m")
                plt.title("Switch points 1 to 7 (4 is our formula, optimal is between 4 and 5)")
                plt.ylabel("I", rotation="horizontal")
                plt.xlabel("S")
                plt.tick_params(labelcolor="w") # don't display numbers (in fact, make them white)
            else:
                plt.text(t1, I1, "$1$", color="k")
                plt.text(tmax, Imax, "$2$", color="k")
                plt.text(t3, I3, "$3$", color="k")
                plt.text(t4, I4, "$4$", color="k")
                plt.text(t5, I5, "$5$", color="k")
                plt.text(t6, I6, "$6$", color="k")
                plt.text(t7, I7, "$7$", color="k")
                Ilist = []
                Idlist = []
                for j in range(len(tlist)):
                    S = Slist[j]
                    Ilist.append(u1*(N-S) / float(y+u1))
                    Idlist.append(Id)
                #plt.plot(tlist, Ilist, color="g")
                plt.plot(tlist, Idlist, color="m")
            plt.yscale("log")
            plt.savefig("switchpoint%slog.png" % xaxis)
            plt.savefig("switchpoint%slog.pdf" % xaxis)
            plt.yscale("linear")
            #plt.ylim(ymin=0, ymax=30)
            plt.savefig("switchpoint%s.png" % xaxis)
            plt.savefig("switchpoint%s.pdf" % xaxis)
