# try to reproduce results from Pollett2001, extended to 2D

import sys
import math
import time
#import scipy as np
import numpy as np
#cy:cimport numpy as np
import cython

#import pyximport; pyximport.install()
import gaussiandict
import Bfunc

def my_sqrt(x):
    if x > 0:
        return math.sqrt(x)
    else:
        return 0

def signed_sqrt(x):
    if x >= 0:
        return math.sqrt(x)
    else:
        return -math.sqrt(-x)

# 4th order R-K method (Kreyszig, 8th Ed (1999) p.958
# If x = (s,i), then we are iteratilvely solving:
# dx = F(x) dt
#cy-
def dx( dt, s, i, BN, y, u1, u2, u3):
#cy+
#cpdef dx(double dt, double s, double i, double BN, double y, double u1, double u2, double u3):
#    cdef double dxs, dxi
#cy.
    dxs = (-BN*s*i + u1*(1-s) + u2*(1-s-i) + u3*i) * dt
    dxi = (BN*s*i - (y+u1)*i - u3*i) * dt
    return (dxs, dxi)

#cy-
def detstep( dt, s, i, BN, y, u1, u2, u3):
#cy+
#cpdef detstep(double dt, double s, double i, double BN, double y, double u1, double u2, double u3):
#    cdef double k1s, k1i, k2s, k2i, k3s, k3i, k4s, k4i, snew, inew #, ds, di
#cy.
    k1s, k1i = dx(dt, s, i, BN, y, u1, u2, u3)
    k2s, k2i = dx(dt, s + 0.5*k1s, i + 0.5*k1i, BN, y, u1, u2, u3)
    k3s, k3i = dx(dt, s + 0.5*k2s, i + 0.5*k2i, BN, y, u1, u2, u3)
    k4s, k4i = dx(dt, s + k3s, i + k3i, BN, y, u1, u2, u3)
    snew = s + (k1s + 2*k2s + 2*k3s + k4s) / 6.0
    inew = i + (k1i + 2*k2i + 2*k3i + k4i) / 6.0
    return (snew, inew)

#cy-
# returns densities: s, i, Fs (ds/dt), Fi (di/dt), 2x2 coveriance array C (capital Sigma in the papers), t
def diffusion_unscaled( N, BN, y, u1, u2, u3, S0, I0,
                        t0=0.0,
                        Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
                        switchcond="",
                        switchset=set(),
                        stickyswitch=1, # once switched, stay switched
                        betamod=0.9,
                        Istop=0, # exactly one of this and Sstop must be set
                        Smax=0, # additional Istop condition
                        verbose=0,
                        op = None,
                        tprintstep=5.0,
                        # usually not used ones go below
                        timeit=0,
                        Sstop=0,
                        comparison="<=", # usually what we want
                        Smin=0, # additional Istop condition
                        Imax=0,  Imin=0, # additional Sstop conditions
                        Cinit = None,
                        ):
#cy+
#cpdef diffusion_unscaled(int N, double BN, double y, double u1, double u2, double u3, double S0, double I0,
#               double t0=0.0,
#               int Bmode=0, double B1=0.0, double Bamp=0.0, double Bperiod=1.0, double Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
#               switchcond="",
#               switchset=set(),
#               int stickyswitch=1,
#               double betamod=0.9,
#               int Istop=0, # exactly one of this and Sstop must be set
#               int Smax=0, # additional Istop condition
#               int verbose=0,
#               op = None,
#               double tprintstep=5.0,
#               # usually not used ones go below
#               int timeit=0,
#               int Sstop=0,
#               comparison="<=", # usually what we want
#               int Smin=0, # additional Istop condition
#               int Imax=0, int Imin=0, # additional Sstop conditions
#               Cinit=None,
#               ):
#    cdef double B, s0, i0, tstep, t, tprint, s, i, smax, smin, imax, imin, sstop, istop
#    cdef double Fs, Fi, Bss, Bsi, Bis, Bii, Gss, Gsi, Gis, Gii, Fiprev, Fsprev, Fiprev2, imins, s_lowest
#    cdef int loop, loop2, exitok, count, bmcount, S, I, in_switchset, len_switchset
#cy.
    if timeit:
        starttime = time.time()
    if not Sstop and not Istop:
        raise Exception("At least one of Sstop and Istop must be set")
    s0 = S0 / float(N)
    i0 = I0 / float(N)
    count = 0
    bmcount = 0
    len_switchset = len(switchset)
    #cy+
    #cdef np.ndarray[np.float64_t, ndim=2] C
    #cdef np.ndarray[np.float64_t, ndim=2] Bu
    #cdef np.ndarray[np.float64_t, ndim=2] BuT
    #cdef np.ndarray[np.float64_t, ndim=2] G
    #cdef np.ndarray[np.float64_t, ndim=2] dC
    #cy.
    # loop2 = correction to be added on 2nd loop
    if comparison.count("<"):
        loop2 = 1
    else:
        loop2 = -1
    for loop in [0, loop2]: # only one loop unless there is an "emergency exit"
        if verbose:
            op.write("diffusion_unscaled loop=%d\n" % loop)
        steps = 0
        tstep = 0.001
        t = t0
        tprint = tprintstep
        s = s0
        i = i0
        s_lowest = s0
        if Cinit is None:
            C = np.zeros((2,2)) # (co)Variance matrix
        else:
            C = np.copy(Cinit)
        cstr = ""
        exitstr = ""
        if Istop:
            istop = (Istop+loop) / float(N)
            cstr = "i " + comparison + " istop"
            if Smax:
                smax = Smax / float(N)
                cstr += " and s <= smax"
            if Smin:
                smin = Smin / float(N)
                cstr += " and s >= smin"
            if verbose>=2 and N<=3000:
                op.write("cstr= " + str(cstr) + "\n")
                op.write("Istop= " + str(Istop) + "\n")
                op.write("istop= " + str(istop) + "\n")
                op.write("smax= " + str(smax) + "\n")
                op.write("Smax= " + str(smax*N) + "\n")
            if comparison.count("<"):
                exitstr = "Fi > 0 and Fiprev <= 0 and imins >= 3"
            else:
                exitstr = "Fi < 0 and Fiprev >= 0 and imins >= 3"
        if Sstop:
            sstop = (Sstop+loop) / float(N)
            if cstr != "":
                cstr += " or "
            cstr += "s " + comparison + " sstop"
            if Imax:
                imax = Imax / float(N)
                cstr += " and i <= imax"
            if Imin:
                imin = Imin / float(N)
                cstr += " and i >= imin"
            if exitstr != "":
                exitstr += " or "
            if comparison.count("<"):
                exitstr += "Fs > 0 and Fsprev <= 0 and imins >= 3"
            else:
                exitstr += "Fs < 0 and Fsprev >= 0 and imins >= 3"
        if comparison.count("<"):
            Fi = 1.0
            Fiprev = 1.0
            Fiprev2 = 1.0
            Fs = 1.0
            Fsprev = 1.0
        else:
            Fi = -1.0
            Fiprev = -1.0
            Fiprev2 = -1.0
            Fs = -1.0
            Fsprev = -1.0
        imins = 0 # count of number of times Fi hits a minimum
        if verbose:
            op.write("starting diffusion at (%g, %g), t=%g\n" % (s, i, t))
            op.write('switchcond = "%s"\n' % switchcond)
            op.write('cstr = "%s"\n' % cstr)
            op.write('exitstr = "%s"\n' % exitstr)
            op.write("BN=%g, betamod=%g\n" % (BN, betamod))
        while 1:
            if Bmode != 0:
                B = Bfunc.Bfunc(Bmode, B1, Bamp, Bperiod, Bphase, t)
            else:
                in_switchset = 0
                if len_switchset > 0:
                    S = int(s*N + 0.5)
                    I = int(i*N + 0.5)
                    if (S,I) in switchset:
                        in_switchset = 1
                if in_switchset or eval(switchcond) or stickyswitch and bmcount>=1:
                    B = BN * betamod
                    bmcount += 1
                else:
                    B = BN
            count += 1
            if verbose>=3:
                op.write("(S,I) = (%g,%g), B=%g\n" % (s*N, i*N, B))
            (s,i) = detstep(tstep, s, i, B, y, u1, u2, u3)
            s_lowest = min(s, s_lowest)
            t += tstep
            steps += 1
            if Fiprev<=0 and Fi>0:
                imins += 1
                if verbose:
                    op.write("imins=%d at (%g,%g) = (%.1f,%.1f)\n" % (imins, s, i, s*N, i*N))
            if eval(cstr):
                exitok = 1
                break
            if eval(exitstr):
                if verbose:
                    op.write("exitstr at (s,i)=(%g,%g), (S,I)=(%.1f,%.1f)\n" % (s, i, s*N, i*N))
                exitok = 0
                break
            Fiprev2 = Fiprev
            Fiprev = Fi
            Fsprev = Fs

            # up-left    (infection)         at rate Bsi
            # right      (s birth / r renew) at rate u1 + u2(1-s-i)
            # down-right (i renew)           at rate u3i
            # down       (recover / i death) at rate (y+u1)i
            # left       (s death)           at rate u1s
            Fs = -B*s*i + u1 - u1*s + u2*(1-s-i) + u3*i
            Fi =  B*s*i - (y+u1)*i - u3*i
            # Bss = dFs/ds = -B*i - u1 - u2
            # Bsi = dFs/di = -B*s - u2 + u3
            # Bis = dFi/ds = B*i
            # Bii = dFi/di = B*s - y - u1 - u3
            Bss = -B*i - u1 - u2
            Bsi = -B*s - u2 + u3
            Bis =  B*i
            Bii =  B*s - y - u1 - u3
            # g11 = (-1)(-1)Bsi + ( 0)( 0)(y+u1)i + (1)(1)(u1 + u2(1-s-i)) + ( 1)( 1)u3 i + (-1)(-1)u1s
            # g12 = (-1)( 1)Bsi + ( 0)(-1)(y+u1)i + (1)(0)(u1 + u2(1-s-i)) + ( 1)(-1)u3 i + (-1)( 0)u1s
            # g21 = ( 1)(-1)Bsi + (-1)( 0)(y+u1)i + (0)(1)(u1 + u2(1-s-i)) + (-1)( 1)u3 i + ( 0)(-1)u1s
            # g22 = ( 1)( 1)Bsi + (-1)(-1)(y+u1)i + (0)(0)(u1 + u2(1-s-i)) + (-1)(-1)u3 i + ( 0)( 0)u1s
            Gss =  B*s*i + u1 + u2*(1-s-i) + u3*i + u1*s
            Gsi = -B*s*i - u3*i
            Gis = -B*s*i - u3*i
            Gii =  B*s*i + (y+u1)*i + u3*i

            Bu = np.matrix([[Bss, Bsi],
                            [Bis, Bii]])
            BuT = np.transpose(Bu)
            G = np.matrix([[Gss, Gsi],
                           [Gis, Gii]])
            dC = (Bu*C + C*BuT + G) * tstep
            C = C + dC
            if verbose >= 2 and t > tprint:
                op.write("t=%g, det=(%.2f, %.2f), sd = (%.2f, %.2f, %.2f), B=%g\n" %
                                 (t, s*N, i*N,
                                  my_sqrt(C[(0,0)]*N),
                                  my_sqrt(C[(1,1)]*N),
                                  signed_sqrt(C[(0,1)]*N), B))
                tprint += tprintstep
        if exitok:
            break
    if not exitok:
        sys.stderr.write("No successful exit from diffusion_unscaled\n")
        sys.stderr.write("exitstr = %s, " % exitstr)
        sys.stderr.write("eval(exitstr) = %s\n" % eval(exitstr))
        return (-1, 0, 0, 0, 0, 0) # correct length tuple with s<0 as a "flag"
    # end of loop. Little to be gained from cython defs after here
    if verbose:
        op.write("exit at (S,I,t) = (%g, %g, %g)\n" % (s*N, i*N, t))
    if verbose:
        op.write("used betamod*B = %g for %d / %d diffusion points\n" % (betamod*BN, bmcount, count))
    if timeit:
        op.write("elapsed time = %g seconds (diffusion.pyx, %d steps)\n" % (time.time() - starttime, steps))
    return (s, i, Fs, Fi, C, t, s_lowest)

def diffusion( N, BN, y, u1, u2, u3, S0, I0,
               t0=0.0,
               Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
               Sd=0, # error in gaussiandict if this is not set to positive value
               switchcond="0", switchset=set(), stickyswitch=1, betamod=0.9,
               Istop=0, # exactly one of this and Sstop must be set
               Smax=0, # additional Istop condition
               verbose=0,
               op = None,
               tprintstep=5.0,
               # usually not used ones go below
               timeit=0,
               Sstop=0,
               comparison="<=", # usually what we want
               Smin=0, # additional Istop condition
               Imax=0,  Imin=0, # additional Sstop conditions
               Cinit = None,
               BN2 = 0, # BN for the 2nd stage (BN in 2nd quadrant, BN2 in 3rd)
               comparison2 = "",
               ):
    if op==None or op=="":
        op = sys.stdout
    if verbose >= 1:
        if Cinit is None:
            op.write("Initial diffusion stddevs at (%.2f, %.2f) are None\n"
                     % (S0, I0))
        else:
            op.write("Initial diffusion stddevs at (%.2f, %.2f) are (%.2f, %.2f, %.2f)\n"
                     % (S0, I0, my_sqrt(Cinit[0,0]*N), my_sqrt(Cinit[1,1]*N), signed_sqrt(Cinit[0,1]*N)))
    if BN2 > 0:
        if comparison2=="":
            comparison2 = comparison
        # use BN up to S=Sstop
        (s1, i1, Fs, Fi, C1, t1, s_lowest) = diffusion_unscaled( N, BN, y, u1, u2, u3, S0, I0,
                                                       t0=t0,
                                                       Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                                       switchcond=switchcond, switchset=switchset, stickyswitch=stickyswitch, betamod=betamod,
                                                       Istop=0, Smax=0, Smin=0,
                                                       Sstop=Sstop, Imax=Imax,  Imin=Imin,
                                                       verbose=verbose,
                                                       op=op,
                                                       tprintstep=tprintstep,
                                                       timeit=timeit,
                                                       comparison=comparison,
                                                       Cinit=Cinit,
                                                       )
        if s1 < 0:
            return ()
        # then use BN2 around to I=Istop
        (s, i, Fs, Fi, C, t2, s_lowest) = diffusion_unscaled( N, BN2, y, u1, u2, u3, s1*N, i1*N,
                                                    t0=t1,
                                                    Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                                    switchcond=switchcond, switchset=switchset, stickyswitch=stickyswitch, betamod=betamod,
                                                    Istop=Istop, Smax=Smax, Smin=Smin,
                                                    Sstop=0, Imax=0,  Imin=0,
                                                    verbose=verbose,
                                                    op=op,
                                                    tprintstep=tprintstep,
                                                    timeit=timeit,
                                                    comparison=comparison,
                                                    Cinit=C1,
                                                    )
        t = t2
    else:
        (s, i, Fs, Fi, C, t, s_lowest) = diffusion_unscaled( N, BN, y, u1, u2, u3, S0, I0,
                                                   t0=t0,
                                                   Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                                   switchcond=switchcond, switchset=switchset, stickyswitch=stickyswitch, betamod=betamod,
                                                   Istop=Istop,
                                                   Smax=Smax,
                                                   verbose=verbose,
                                                   op=op,
                                                   tprintstep=tprintstep,
                                                   # usually not used ones go below
                                                   timeit=timeit,
                                                   Sstop=Sstop,
                                                   comparison=comparison,
                                                   Smin=Smin,
                                                   Imax=Imax,  Imin=Imin,
                                                   Cinit=Cinit,
                                                   )
    if s < 0:
        return ()
    # use final values of Fs and Fi
    if Istop:
        hitvar0 = C[0,0] - (Fs/Fi) * C[1,1]
        hitvar1 = C[0,0] + (Fs/Fi)**2 * C[1,1]
        hitvar  = C[0,0] + (Fs/Fi)**2 * C[1,1] - 2*(Fs/Fi)*C[0,1]
    else:
        hitvar0 = C[1,1] - (Fi/Fs) * C[0,0]
        hitvar1 = C[1,1] + (Fi/Fs)**2 * C[0,0]
        hitvar  = C[1,1] + (Fi/Fs)**2 * C[0,0] - 2*(Fi/Fs)*C[0,1]
    hitsd0 = my_sqrt(hitvar0) / math.sqrt(N)
    hitsd1 = my_sqrt(hitvar1) / math.sqrt(N)
    hitsd  = my_sqrt(hitvar)  / math.sqrt(N)
    # approximate sd in i, ignoring covariance
    isd = math.sqrt(C[1,1]) / math.sqrt(N)
    # approximate sd in t, ignoring covariance
    if comparison.count("<"):
        tsd = isd / (-Fi) # Fi = di/dt. Negative because i is falling (usually)
    else:
        tsd = isd / Fi # Fi = di/dt
    if verbose >= 1:
        if Istop:
            stopstr = "Istop"
        else:
            stopstr = "Sstop"
        op.write("%s diffusion stddevs at (%.2f, %.2f) are (%.2f, %.2f, %.2f) -> hitsd= %.1f, tsd= %g\n"
                         % (stopstr, s*N, i*N, my_sqrt(C[0,0]*N), my_sqrt(C[1,1]*N), signed_sqrt(C[0,1]*N), hitsd*N, tsd))
        if Istop:
            # discretise
            Sdict = gaussiandict.gaussiandict(op, s*N, hitsd*N, 0, Sd-1, rescale=1)
            dmean = 0.0
            dmean2 = 0.0
            dtot = 0.0
            for key in Sdict:
                dtot   += Sdict[key]
                dmean  += Sdict[key] * key
                dmean2 += Sdict[key] * key*key
            dmean = dmean / dtot
            dmean2 = dmean2 / dtot
            dsd = my_sqrt(dmean2 - dmean*dmean)
            op.write("After discretisation (of dtot %.3f ),            mean %.1f and sd %.1f\n" % (dtot, dmean, dsd))
    if Istop:
        return (s*N, hitsd*N, t, tsd, s_lowest*N)
    else:
        return (i*N, hitsd*N, t, tsd, s_lowest*N)
