# return 1st max, 1st min and 2nd max of a deterministic SIR simulation

import sys
import math

import sir_stable
import Bfunc

def maxratio(dS, S, dI, I, dR, R):
    ratio = 0
    if S > 0:
        ratio = max( abs(dS/float(S)), ratio)
    if I > 0:
        ratio = max( abs(dI/float(I)), ratio)
    if R > 0:
        ratio = max( abs(dR/float(R)), ratio)
    return ratio

def dx(N, S, I, R, BonN, y, u1, u2, u3, dt):
    Rdeaths = R * (u1+u2)*dt
    Ideaths = I * u3*dt
    Sbirths = ((N-S)*u1 + R*u2) *dt
    Inew = I * BonN * S * dt
    Rnew = (y+u1) * I*dt
    dS = -Inew + Sbirths + Ideaths
    dI = Inew - Rnew - Ideaths
    dR = Rnew - Rdeaths
    return (dS, dI, dR)

# 4th order R-K method (Kreyszig, 8th Ed (1999) p.958
def detstep(N, S, I, R, BonN, y, u1, u2, u3, dt):
    k1S, k1I, k1R = dx(N, S, I, R, BonN, y, u1, u2, u3, dt)
    k2S, k2I, k2R = dx(N, S+0.5*k1S, I+0.5*k1I, R+0.5*k1R, BonN, y, u1, u2, u3, dt)
    k3S, k3I, k3R = dx(N, S+0.5*k2S, I+0.5*k2I, R+0.5*k2R, BonN, y, u1, u2, u3, dt)
    k4S, k4I, k4R = dx(N, S+k3S, I+k3I, R+k3R, BonN, y, u1, u2, u3, dt)
    dS = (k1S + 2*k2S + 2*k3S + k4S) / 6.0
    dI = (k1I + 2*k2I + 2*k3I + k4I) / 6.0
    dR = (k1R + 2*k2R + 2*k3R + k4R) / 6.0
    return (dS, dI, dR)

# p0, if we pretend it's a 1-D problem (i.e. S constant)
def p0_1d(S, I, BonN, y, Istable):
    r = BonN*S/float(y)
    if r==1:
        return (Istable - I) / Istable
    else:
        if I > Istable:
            return 0
        try:
            big = r**Istable
        except OverflowError:
            return 1.0 # r**Istable dwarfs other terms
        try:
            return (big - r**I) / (big - 1)
        except OverflowError:
            raise OverflowError("S, I, BonN, y, Istable = %s, %s, %s, %s, %s" % (S, I, BonN, y, Istable))

# round the S and I of an (S,I,t) tuple if forceint
def SIt_round(S, I, t, forceint):
    if forceint:
        return ( int(S + 0.5), int(I + 0.5), t)
    else:
        return ( S, I, t)

# run to an exit point
# This was created because sirdetpeaks was getting unwieldy
# Returns floats; it is up to calling program to deal with
def sirdetrun(op=sys.stdout, N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, S0=-1, t0=0,
              tstep=0.001, tstepfixed=False, verbose=0,
              Istop=None, Sstop=None, dir=0, # 1 for rising, -1 for falling, 0 for either
              condstop="", # a string to evaluate
              Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
              forceint=False,
              retlist=0, retdt=0, Rinit=-1,
              ):
    if not Istop and not Sstop and not condstop:
        raise Exception(("Either Istop or Sstop or condstop must be set: (S0,I0)=(%g,%g)" % (S0, I0)))
    t = t0
    I = I0
    if S0 < 0:
        S = N - I0
    else:
        S = S0
    BonN = BN / float(N)
    Bprev = 0
    if Bmode==0:
        (Sd, Id) = sir_stable.stable(N, BN, y, u1, u2, u3)
    if Rinit >= 0:
        R = Rinit
    else:
        R = N - S - I
    Ichanges = 0
    dt = tstep # initial try
    dmax = 0.0 # purely for the first write; init value not actually used
    if retlist:
        Slist = [S]
        Ilist = [I]
        tlist = [t]
    if verbose >= 2:
        op.write("starting at (%g,%g)\n" % (S, I))
        nextprint = t + 0.01
    while 1:
        if Bmode != 0:
            B = Bfunc.Bfunc(Bmode, B1, Bamp, Bperiod, Bphase, t)
            if verbose and B!=Bprev and Bfunc.Bedges(Bmode):
                op.write("(%.1f, %.1f, %g): B changing to %g\n" % (S, I, t, B))
            Bprev = B
            BonN = B / float(N)
            (Sd, Id) = sir_stable.stable(N, B, y, u1, u2, u3)
        (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
        if verbose >=3 and t > nextprint or verbose >= 4:
            op.write("(%s,%s,%s),\tdI=%g dt=%s\tBfunc(%s, %s, %s, %s, %s, %s), B=%s, (Sd,Id)=(%d,%d)\n"
                     % (S,I,t, dI, dt, Bmode, B1, Bamp, Bperiod, Bphase, t, BonN*N, Sd, Id))
            nextprint += 0.01
        if not tstepfixed:
            if verbose >= 4:
                op.write("(S,I,R)=(%s, %s, %s), (dS,dI,dR)=(%s, %s, %s)\n"
                                 % (S,I,R, dS, dI, dR))
            dmax = maxratio(dS, S, dI, I, dR, R)
            if dmax > 0.01:
                dt = dt * 0.01 / float(dmax)
                (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
            if dmax < 0.0001:
                dt = dt * 0.0001 / float(dmax)
                (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
        if Istop:
            if (   I+dI >= Istop and I<Istop  and dir in [0, 1] # rising
                or I+dI <  Istop and I>=Istop and dir in [0, -1]): # falling
                if dI==0.0:
                    Sinterp = S
                else:
                    Sinterp = S + dS * (Istop-I) / float(dI)
                tup = SIt_round(Sinterp, Istop, t, forceint)
                break
        if Sstop:
            if (   S+dS >= Sstop and S<Sstop  and dir in [0, 1] # rising
                or S+dS <  Sstop and S>=Sstop and dir in [0, -1]): # falling
                if dS==0.0:
                    Iinterp = I
                else:
                    Iinterp = I + dI * (Sstop-S) / float(dS)
                tup = SIt_round(Sstop, Iinterp, t, forceint)
                break
        if condstop and eval(condstop):
            tup = SIt_round(S, I, t, forceint)
            break
        if (   I+dI >= Istop and I<Istop
            or I+dI <  Istop and I>=Istop):
            Ichanges += 1
            if Ichanges >= 4:
                raise Exception("Ichanges=4, should have exited by now. Something is wrong. (S,I,t)=(%s,%s,%s)" % (S,I,t))
        S = S+dS
        I = I+dI
        #if Nfixed:
        #    R = N - S - I
        #else:
        R = R + dR
        t = t + dt
        if retlist:
            Slist.append(S)
            Ilist.append(I)
            tlist.append(t)
        # duplicate
        #if verbose >= 3:
        #    op.write("(%g, %g, %g)\n" % (S, I, t))
    if retlist:
        if retdt:
            return [tup] + [[Slist, Ilist, tlist]] + [dt, R, dS, dI, dR]
        else:
            return [tup] + [[Slist, Ilist, tlist]]
    else:
        if retdt:
            return tup + (dt, R, dS, dI, dR)
        else:
            return tup

# Runs until 2nd maximum (max2) or I>=Imax
# 12-Sep-2014: make call by keyword, and *change Nfixed default* from False to None
# "point" is as defined for function defpoint, as an exit condition,
# but only values 1 (first Istable), 2 (first Imax), 5 (first Imin) are supported
def sirdetpeaks(N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, S0=-1, threshold=0, forceint=False,
                tstep=0.001, tstepfixed=False, verbose=0, point=0):
    # forceint forces S and I to nearest integer
    # (S,I,t) tuples
    BonN = BN / float(N)
    (Se, Ie) = sir_stable.stable_float(N, BN, y, u1, u2, u3)
    max1 = (0.0, 0.0, -1.0)
    min1 = (0.0, 0.0, -1.0)
    max2 = (0.0, 0.0, -1.0)
    t = 0.0
    I = I0
    if S0 < 0:
        S = N - I0
    else:
        S = S0
    R = 0
    dt = tstep # initial try
    if verbose >= 2:
        sys.stdout.write("dt=%g\n" % dt)
    xover = [] # list of (S,I,t) tuples at which i=threshold
    if verbose >= 2:
        sys.stdout.write("starting at (%g,%g)\n" % (S, I))
    count = 0
    nextprint = 1
    while I > 0:
        (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
        if not tstepfixed:
            dmax = maxratio(dS, S, dI, I, dR, R)
            if dmax > 0.01:
                dt = dt * 0.01 / float(dmax)
                if verbose >= 2:
                    sys.stdout.write("dt=%g\n" % dt)
                (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
            if dmax < 0.0001:
                dt = dt * 0.0001 / float(dmax)
                if verbose >= 2:
                    sys.stdout.write("dt=%g\n" % dt)
                (dS, dI, dR) = detstep(N, S, I, R, BonN, y, u1, u2, u3, dt)
        # weighted. Note Ie is type float.
        nom   = p0_1d(S+1, I, BonN, y, Ie) - p0_1d(S, I, BonN, y, Ie)
        denom = p0_1d(S, I+1, BonN, y, Ie) - p0_1d(S, I, BonN, y, Ie)
        if nom==denom:
            weight = 1.0
        else:
            try:
                weight = (  (p0_1d(S+1, I, BonN, y, Ie) - p0_1d(S, I, BonN, y, Ie))
                          / (p0_1d(S, I+1, BonN, y, Ie) - p0_1d(S, I, BonN, y, Ie)))
            except ZeroDivisionError:
                weight = 1.0
                #raise Exception("ZeroDivisionError at S=%d, I=%d, BonN=%s y=%s, Istable=%s\n(%s - %s) / (%s - %s) = %s / %s"
                #                  % (S, I, BonN, y, Ie,
                #                     p0_1d(S+1, I, BonN, y, Ie), p0_1d(S, I, BonN, y, Ie),
                #                     p0_1d(S, I+1, BonN, y, Ie), p0_1d(S, I, BonN, y, Ie),
                #                     nom, denom))
        if (I < threshold and I+dI >= threshold) or (I >= threshold and I+dI < threshold):
            txover = t + dt*(threshold-I) / float(dI)
            sxover = S + dS*(threshold-I) / float(dI)
            xover.append(SIt_round(sxover, threshold, txover, forceint)) # (S,I,t) tuple
        Sprev = S
        Iprev = I
        S = S+dS
        I = I+dI
        R = N - S - I
        count += 1
        if verbose >= 3 or verbose>=2 and count>=nextprint:
            sys.stdout.write("%d, %g: (%s,%s) to (%s,%s) step=(%g,%g)\n"
                             % (count, t, repr(Sprev), repr(Iprev), repr(S), repr(I), dS, dI))
        if count >= nextprint:
            nextprint = 2*count
        # these are all (S,I,t) tuples
        if max1[2] < 0 and t>0 and dI_prev >= 0 and dI < 0:
            max1 = SIt_round(S-dS, I-dI, t + dt/2.0, forceint)
            if verbose >= 2:
                sys.stdout.write("max1 %g: (%g,%g) -> (%g,%g)\n" % (t, S-dS, I-dI, S, I))
            if point==2:
                break
        elif min1[2] < 0 and t>0 and dI_prev <= 0 and dI > 0:
            # Note min1 can be before max1
            min1 = SIt_round(S-dS, I-dI, t + dt/2.0, forceint)
            if verbose >= 2:
                sys.stdout.write("min1 %g: (%g,%g) -> (%g,%g)\n" % (t, S-dS, I-dI, S, I))
            if point==5:
                break
        elif min1[2] > 0 and max2[2] < 0 and dI < 0: # and dI_prev>=0  -> infinite loop. Not sure why.
            max2 = SIt_round(S-dS, I-dI, t + dt/2.0, forceint)
            if verbose >= 2:
                sys.stdout.write("max2 %g: (%g,%g) -> (%g,%g)\n" % (t, S-dS, I-dI, S, I))
            break
        elif point==1 and I>=Ie:
            break
        # hack to make sure it always terminates
        elif I==Iprev and S==Sprev:
            sys.stdout.write("[Stalled at (%g,%g)]" % (S, I))
            break
        #elif abs(dI)<1e-12 and abs(dS)<1e-12:
        #    break
        t = t + dt
        dI_prev = dI
    if verbose >= 2:
        sys.stdout.write("ending at (%g,%g)\n" % (S, I))
    #sys.stdout.write("max1=%f, min1=%f, max2=%f, t=%f\n", max1, min1, max2, t)
    lenxover = len(xover)
    for j in range(lenxover, 3):
        xover.append((Se,Ie,0.0))
    return (max1, min1, max2, xover)

# returns a single (S,I,t) tuple based on the parameter
# 0 = start
# 1 = first (rising) I==Istablef (will not exist if I0 >= Istablef)
# 2 = first Imax
# 3 = first falling I==Istablef
# 4 = I = halfway between points 3 and 5  (rounded to nearest I)
# 5 = Imin of first trough
# 6 = 2nd rising I==Istablef (or first if I0 >= Istablef)
# 7 = 2nd Imax
# int forces S and I to nearest integer
def detpoint(N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, S0=-1,
             point=0, forceint=False, tstep=0.001, tstepfixed=False, verbose=0):
    # see ../lib/sirplot.py for derivations
    BonN = BN / float(N)
    (Se, Ie) = sir_stable.stable_float(N, BN, y, u1, u2, u3)
    (max1, min1, max2, xover) = sirdetpeaks(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                            threshold=Ie, forceint=forceint,
                                            tstep=tstep, tstepfixed=tstepfixed, point=point, verbose=verbose)
    if point==0:
        return (N-I0, I0, 0)
    elif point==1:
        if I0 >= Ie:
            sys.stdout.write("no first xover because I0=%d > Ie=%.1f" % (I0, Ie))
            # returns None. Will probably raise an error in the calling program.
        else:
            return xover[0]
    elif point==2:
        return max1
    elif point==3:
        if I0 >= Ie:
            return xover[0]
        else:
            return xover[1]
    elif point==4:
        threshold2 = int((Ie + min1[1]) / 2.0 + 0.5) # I: average of Ie and Imin (rounded to nearest I)
        (max1, min1, max2, xover) = sirdetpeaks(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                threshold=threshold2, forceint=forceint,
                                                tstep=tstep, tstepfixed=tstepfixed, point=4, verbose=verbose)
        if I0 >= threshold2:
            return xover[0]
        else:
            return xover[1]
    elif point==5:
        return min1
    elif point==6:
        if I0 >= Ie:
            return xover[1]
        else:
            return xover[2]
    elif point==7:
        return max2

# find the det point at I=Istable-1, i.e. starting point for sims
def sweepstartpoint_float(N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, S0=-1,
                    tstep=0.001, tstepfixed=True, verbose=0):
    (Sstable, Istable) = sir_stable.stable(N, BN, y, u1, u2, u3)
    (S, I, t) = sirdetrun(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0, 
                          Istop=Istable-1, dir=-1,
                          tstep=tstep, tstepfixed=tstepfixed, verbose=verbose)
    return (S, Istable)

def sweepstartpoint(N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, S0=-1, 
                    tstep=0.001, tstepfixed=True, verbose=0):
    (Sstable, Istable) = sir_stable.stable(N, BN, y, u1, u2, u3)
    (S, I, t) = sirdetrun(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0, 
                          Istop=Istable-1, dir=-1,
                          tstep=tstep, tstepfixed=tstepfixed, verbose=verbose)
    return (int(S+0.5), Istable-1)

# Computes the things we want for a 2-bounds test
# calls depoints a couple of times with different point values
def detpoints_for_2bounds(N=None, BN=1, y=None, u1=0, u2=0, u3=0, I0=1, 
                          Imin=0, Imax=-1, startpoint=-1, Imax2=-1, S0=-1,
                          monitor_S=False, verbose=0):
    (Sstable, Istable) = sir_stable.stable(N, BN, y, u1, u2, u3)
    if Imax < 0:
        Imax = Istable
    if Imax2 < 0:
        Imax2 = max(int(math.ceil( detpoint(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, point=7)[1])), 2*Istable)
    if startpoint >= 0: # Note S0 is always ignored in this case
        (Sstart, Istart, tignore, dignore, dignore2) = detpoint(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, point=startpoint, forceint=True)
    else:
        Istart = I0
        if S0 < 0:
            Sstart = N - Istart
        else:
            Sstart = S0
    if verbose > 0:
        sys.stdout.write("(Sstart,Istart)=(%s,%s), Imin=%s, Imax=%s" % (Sstart, Istart, Imin, Imax))
        if monitor_S:
            sys.stdout.write(", Sstable=%s, Imax2=%s" % (Sstable, Imax2))
        sys.stdout.write("\n")
    return (Sstart, Istart, Sstable, Imax, Imax2)
